import re
import time

import numpy as np
from bidict import frozenbidict
from pint.errors import UndefinedUnitError

import kamzik3
from kamzik3 import DeviceError, CommandFormatException, DeviceUnitError
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.devices.general.deviceScanner import DeviceScanner
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

""" Example of yaml configuration
Smaract0: &Smaract0 !Device:kamzik3.devices.deviceSmaractMcs1.DeviceSmaractMcs2
    device_id: Smaract0
    host: 192.168.83.91
Smaract0_axis_0: &Smaract0_axis_0 !Device:kamzik3.devices.deviceSmaractMcs1.DeviceSmaractMcs1Channe2
    device: *Smaract0
    channel: 0
    device_id: MotorAxis0
    config:
      attributes:
        !!python/tuple [Sensor type, Value]: 1
        !!python/tuple [Position +limit, Value]: 15mm
        !!python/tuple [Position -limit, Value]: -15mm
        !!python/tuple [Position, Tolerance]: [10nm, 10nm]
"""

"""
Positioner codes:
SL...S1SS 300 Linear positioner, single piezo element
SL...D1SS 301 Linear positioner, double piezo element
SL...S1SC1 303 Linear positioner, single piezo element
SL...D1SC1 304 Linear positioner, double piezo element
SL...D1SC2 307 Linear positioner, double piezo element
SR...S1S5S 309 Small rotary positioner S single mark
SR...S1S6S 312 Rotary positioner S single mark
SR...D1S6S 313 Rotary positioner, double piezo element
SR...D1S7S 316 Rotary positioner, double piezo element
SR...T1S8S 320 Large rotary positioner S single mark
SG...D1S1S 325 Goniometer, 60.5mm radius, double
SG...D1S2S 328 Goniometer, 77.5mm radius, double
SL...S1LE 342 Linear positioner, single piezo element
SL...D1LE 343 Linear positioner, double piezo element
SL...S1LS 345 Linear positioner, single piezo element
SL...D1LS 346 Linear positioner, double piezo element
SL...S1LC1 348 Linear positioner, single piezo element
SL...D1LC1 349 Linear positioner, double piezo element
SR...S1L2S 354 Rotary positioner L single mark
SR...D1L2S 355 Rotary positioner, double piezo element
SL...S1ME 357 Linear positioner, single piezo element
SL...D1ME 358 Linear positioner, double piezo element
SL...S1P1E 360 Linear positioner, double piezo element
SL...D1P1E 361 Linear positioner, double piezo element
SG...S1M1E 363 Goniometer, 60.5mm radius, single
SG...S1M2E 366 Goniometer, 77.5mm radius, single
SG...D1L1S 381 Goniometer, 60.5mm radius, double
SG...D1L2S 383 Goniometer, 77.5mm radius, double
SG...D1L1E 387 Goniometer, 60.5mm radius, double
SG...D1L2E 389 Goniometer, 77.5mm radius, double
SH...A1SS 395 High load table, single piezo element S single mark
"""


class DeviceSmaractMcs2(DeviceSocket):
    response_timeout = 3000
    terminator = b"\r\n"
    error_pattern = re.compile('([-0-9]+),\"([\w ]+)"')
    sensor_modes = frozenbidict({"Disabled": 0, "Enabled": 1, "Power safe": 2})
    move_modes = frozenbidict({"Absolute": 0, "Relative": 1, "Scan absolute": 2, "Scan relative": 3, "Step": 4})
    safe_direction_modes = frozenbidict({"Forward": 0, "Backward": 1})
    emergency_stop_modes = frozenbidict({"Normal": 0, "Restricted": 1, "Auto release": 2})

    def __init__(self, host, port=55551, device_id=None, config=None):
        super(DeviceSmaractMcs2, self).__init__(host, port, device_id, config)

    def _init_attributes(self):
        super(DeviceSmaractMcs2, self)._init_attributes()
        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_CHANNELS, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_MODULES, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_HAND_TOOL, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_MOVEMENT_LOCKED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_INTERNAL_COMM_FAILURE, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_STREAMING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_EMERGENCY_STOP, default_type=self.emergency_stop_modes.keys(),
                              set_function=self.set_emergency_stop_mode)

    def command(self, command, callback=None, with_token=False, returning=True):
        if not self.valid_command_format(command):
            raise CommandFormatException(u"Command '{}' form is invalid".format(command))

        returning = True
        command += ":SYST:ERR:NEXT?\r\n"
        return super(DeviceSmaractMcs2, self).command(command, callback, with_token, returning)

    def poll_command(self, command, interval):
        if not self.valid_command_format(command):
            raise CommandFormatException(u"Command '{}' form is invalid".format(command))

        command += ":SYST:ERR:NEXT?\r\n"
        super(DeviceSmaractMcs2, self).poll_command(command, interval)

    def remove_poll_command(self, command, interval):
        if not self.valid_command_format(command):
            raise CommandFormatException(u"Command '{}' form is invalid".format(command))

        command += ":SYST:ERR:NEXT?\r\n"
        super(DeviceSmaractMcs2, self).remove_poll_command(command, interval)

    def handle_configuration(self):

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        # Check for possible errors
        self.get_errors_count()
        # Get number of channels
        self.command(u":DEV:NOCH?\r\n")
        # Get serial number
        self.command(u":DEV:SNUM?\r\n")
        # Get number of modules
        self.command(u":DEV:NOBM?\r\n", callback=_finish_configuration)
        # Get emergency stop mode
        self.command(u":DEV:EST:MODE?\r\n")
        self.start_polling()

    def handle_configuration_event(self):
        try:
            start_at = time.time()
            self.set_status(STATUS_CONFIGURING)
            self.handle_configuration()
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))
        except DeviceError:
            self.logger.exception(u"Error during configuration")

    def valid_command_format(self, command):
        if command is not None and command[-2:] == "\r\n":
            return True
        else:
            return False

    def handle_readout(self, readout_buffer):
        is_error_message = self.error_pattern.match(readout_buffer[0])
        command, output, callback, token = super(DeviceSmaractMcs2, self).handle_readout(readout_buffer)
        if is_error_message is not None:
            if command == ":SYST:ERR:NEXT?\r\n" and int(is_error_message.group(1)) != 0:
                self.logger.error(readout_buffer[0])
            elif int(is_error_message.group(1)) != 0:
                self.logger.error(readout_buffer[0])

            if callback is not None:
                self.handle_readout_callback(callback, command, readout_buffer)
            if token:
                kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)
                for observer in self._observers[:]:
                    if isinstance(observer, DeviceSmaractMcs2Channel):
                        kamzik3.session.publisher.push_message(observer.device_id, [command, readout_buffer], token)
            return
        else:
            command = command.replace(":SYST:ERR:NEXT?\r\n", "")
            self.commands_buffer.appendleft(((":SYST:ERR:NEXT?\r\n", None, None, True), time.time()))

        command = command.strip()

        if command == ":DEV:NOCH?":
            self.set_raw_value(ATTR_CHANNELS, int(output))
        elif command == ":DEV:SNUM?":
            self.set_raw_value(ATTR_SERIAL_NUMBER, output.replace('"', ""))
        elif command == ":DEV:NOBM?":
            self.set_raw_value(ATTR_MODULES, int(output))
            for i in range(int(output)):
                module_name = "Module{}".format(i)
                self.create_attribute(ATTR_STATE, module_name, readonly=True, default_type=np.uint16)
                self.create_attribute(ATTR_TEMPERATURE, module_name, readonly=True, default_type=np.float32, decimals=2,
                                      unit="degC")
                self.create_attribute(ATTR_SENSOR_PRESENT, module_name, readonly=True, default_type=np.bool)
                self.create_attribute(ATTR_BOOSTER_PRESENT, module_name, readonly=True, default_type=np.bool)
                self.create_attribute(ATTR_ADJUSTMENT_ACTIVE, module_name, readonly=True, default_type=np.bool)
                self.create_attribute(ATTR_IOM_PRESENT, module_name, readonly=True, default_type=np.bool)
                self.create_attribute(ATTR_INTERNAL_COMM_FAILURE, module_name, readonly=True, default_type=np.bool)
                self.create_attribute(ATTR_HIGH_VOLTAGE_FAILURE, module_name, readonly=True, default_type=np.bool)
                self.create_attribute(ATTR_HIGH_VOLTAGE_OVERLOAD, module_name, readonly=True, default_type=np.bool)
                self.create_attribute(ATTR_OVER_TEMPERATURE, module_name, readonly=True, default_type=np.bool)

                self.poll_command(":MOD{}:STAT?\r\n".format(i), 2000)
                self.poll_command(":MOD{}:TEMP?\r\n".format(i), 2000)
        elif command == ":DEV:STAT?":
            output = int(output)
            self.set_raw_value(ATTR_STATE, output)
            self.set_raw_value(ATTR_HAND_TOOL, bool(output % 0x0001))
            self.set_raw_value(ATTR_MOVEMENT_LOCKED, bool(output % 0x0002))
            self.set_raw_value(ATTR_INTERNAL_COMM_FAILURE, bool(output % 0x0100))
            self.set_raw_value(ATTR_STREAMING, bool(output % 0x1000))
        elif command == ":DEV:EST:MODE?":
            output = int(output)
            self.set_raw_value(ATTR_EMERGENCY_STOP, self.emergency_stop_modes.inv[output])
        elif command == ":SYST:ERR:COUN?":
            if int(output) > 0:
                self.get_next_error()
        elif command == ":SYST:ERR:NEXT?":
            self.logger.error(output)
            self.get_errors_count()
        elif command[:4] == ":MOD":
            _, module, attribute = command.split(":")
            module_name = "Module{}".format(module[3:])
            if attribute == "TEMP?":
                self.set_raw_value((module_name, ATTR_TEMPERATURE), float(output))
            elif attribute == "STAT?":
                output = int(output)
                self.set_raw_value((module_name, ATTR_STATE), output)
                self.set_raw_value((module_name, ATTR_SENSOR_PRESENT), bool(output & 0x0001))
                self.set_raw_value((module_name, ATTR_BOOSTER_PRESENT), bool(output & 0x0002))
                self.set_raw_value((module_name, ATTR_ADJUSTMENT_ACTIVE), bool(output & 0x0004))
                self.set_raw_value((module_name, ATTR_IOM_PRESENT), bool(output & 0x0008))
                self.set_raw_value((module_name, ATTR_INTERNAL_COMM_FAILURE), bool(output & 0x0100))
                self.set_raw_value((module_name, ATTR_HIGH_VOLTAGE_FAILURE), bool(output & 0x1000))
                self.set_raw_value((module_name, ATTR_HIGH_VOLTAGE_OVERLOAD), bool(output & 0x2000))
                self.set_raw_value((module_name, ATTR_OVER_TEMPERATURE), bool(output & 0x4000))
        elif command[:5] == ":CHAN":
            chunks = command.split(":")
            channel, attribute = chunks[1], chunks[-1]
            channel = int(channel[4:])
            if attribute == "STAT?":
                attribute, output = ATTR_STATE, int(output)
            elif attribute == "CURR?":
                attribute, output = ATTR_POSITION, int(output)
            elif attribute == "TARG?":
                attribute, output = ATTR_TARGET_POSITION, int(output)
            elif attribute == "TEMP?":
                attribute, output = ATTR_TEMPERATURE, float(output)
            elif attribute == "UNIT?":
                attribute, output = ATTR_ENCODER_BASE_UNIT, int(output)
            elif attribute == "RES?":
                attribute, output = ATTR_ENCODER_RESOLUTION, int(output)
            elif attribute == "NAME?":
                attribute, output = ATTR_POSITIONER_TYPE, output.replace('"', "")
            elif attribute == "CODE?":
                attribute, output = ATTR_POSITIONER_CODE, int(output)
            elif attribute == "MIN?":
                attribute, output = ATTR_POSITION_LOWER_LIMIT, int(output)
            elif attribute == "MAX?":
                attribute, output = ATTR_POSITION_UPPER_LIMIT, int(output)
            elif attribute == "FREQ?":
                attribute, output = ATTR_STEP_FREQUENCY, int(output)
            elif attribute == "AMPL?":
                attribute, output = ATTR_STEP_AMPLITUDE, int(output)
            elif attribute == "VEL?":
                attribute, output = ATTR_VELOCITY, int(output)
            elif attribute == "HOLD?":
                attribute, output = ATTR_HOLDING_TIME, int(output)
            elif attribute == "ACC?":
                attribute, output = ATTR_ACCELERATION, int(output)
            elif attribute == "MODE?":
                attribute, output = ATTR_SENSOR_MODE, self.sensor_modes.inv[int(output)]
            elif attribute == "MMOD?":
                attribute, output = ATTR_MOVE_MODE, self.move_modes.inv[int(output)]
            elif attribute == "SDIR?":
                attribute, output = ATTR_SAFE_DIRECTION, self.safe_direction_modes.inv[int(output)]
            else:
                return
            self.notify((channel, attribute), output)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)
            for observer in self._observers[:]:
                if isinstance(observer, DeviceSmaractMcs2Channel):
                    kamzik3.session.publisher.push_message(observer.device_id, [command, readout_buffer], token)

    def start_polling(self):
        super(DeviceSmaractMcs2, self).start_polling()
        self.poll_command(":DEV:STAT?\r\n", 1000)

    def get_errors_count(self):
        self.command(u":SYST:ERR:COUN?\r\n")

    def get_next_error(self):
        self.command(u":SYST:ERR:NEXT?\r\n")

    def set_emergency_stop_mode(self, value, callback=None):
        self.logger.info(u"Set emergency stop mode to {}".format(value))
        value = self.emergency_stop_modes[value]
        # Set this mode as a default
        self.command(u":DEV:EST:MODE:DEF {}\r\n".format(value), callback, with_token=True)
        return self.command(u":DEV:EST:MODE {}\r\n".format(value), callback, with_token=True)


class DeviceSmaractMcs2Channel(DeviceChannel):
    sensor_modes = DeviceSmaractMcs2.sensor_modes
    move_modes = DeviceSmaractMcs2.move_modes
    safe_direction_modes = DeviceSmaractMcs2.safe_direction_modes

    def _init_attributes(self):
        super(DeviceSmaractMcs2Channel, self)._init_attributes()
        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_TEMPERATURE, readonly=True, default_type=np.float16, decimals=2, unit="degC")
        self.create_attribute(ATTR_MOVING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_POSITION, readonly=False, default_type=np.int64, set_function=self._move_absolute,
                              set_value_when_set_function=False)
        self.create_attribute(ATTR_TARGET_POSITION, default_type=np.int64, set_function=self._move_absolute,
                              set_value_when_set_function=False)
        self.create_attribute(ATTR_CLOSED_LOOP_ACTIVE, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_CALIBRATING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_REFERENCING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_HOLDING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_MOVE_DELAYED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_SENSOR_PRESENT, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_CALIBRATED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_REFERENCED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_END_STOP_REACHED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_RANGE_LIMIT_REACHED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_FOLLOWING_LIMIT_REACHED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_MOVEMENT_FAILED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_STREAMING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_OVER_TEMPERATURE, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_INDEX_MARK, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_ENCODER_BASE_UNIT, readonly=True)
        self.create_attribute(ATTR_ENCODER_RESOLUTION, readonly=True, default_type=np.int8)
        self.create_attribute(ATTR_POSITIONER_CODE, readonly=False, default_type=np.int8, min_value=0, max_value=500,
                              set_function=self.set_positioner_code)
        self.create_attribute(ATTR_POSITIONER_TYPE, readonly=True)
        self.create_attribute(ATTR_POSITION_LOWER_LIMIT, default_type=np.int64,
                              set_function=self.set_lower_position_limit)
        self.create_attribute(ATTR_POSITION_UPPER_LIMIT, default_type=np.int64,
                              set_function=self.set_upper_position_limit)
        self.create_attribute(ATTR_STEP_FREQUENCY, default_type=np.uint32, min_value=1, max_value=20e3, unit="Hz",
                              set_function=self.set_step_frequency)
        self.create_attribute(ATTR_STEP_AMPLITUDE, default_type=np.uint32, min_value=1, max_value=65535,
                              set_function=self.set_step_amplitude)
        self.create_attribute(ATTR_ACCELERATION, default_type=np.int64, min_value=0, max_value=100e9,
                              set_function=self.set_acceleration)
        self.create_attribute(ATTR_VELOCITY, default_type=np.int64, min_value=0, max_value=10e12,
                              set_function=self.set_velocity)
        self.create_attribute(ATTR_SENSOR_MODE, default_type=self.sensor_modes.keys(),
                              set_function=self.set_sensor_mode)
        self.create_attribute(ATTR_MOVE_MODE, default_type=self.move_modes.keys(), set_function=self.set_move_mode)
        self.create_attribute(ATTR_SAFE_DIRECTION, default_type=self.safe_direction_modes.keys(),
                              set_function=self.set_safe_direction)
        self.create_attribute(ATTR_HOLDING_TIME, default_type=np.int32, min_value=-1, unit="ms",
                              set_function=self.set_holding_time)
        self.create_attribute(ATTR_CLOSE_LOOP_UNIT, display=False)
        self.create_attribute(ATTR_CLOSE_LOOP_FACTOR, default_value=[1, 1], default_type=TYPE_LIST, display=False)
        self.create_attribute(ATTR_MOVE_METHOD, default_value=CLOSE_LOOP, default_type=[OPEN_LOOP, CLOSE_LOOP],
                              set_function=self.set_move_method)

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        self.get_sensor_mode()
        self.get_encoder_unit()
        self.get_position_limit_range()
        self.get_positioner_configuration()
        self.get_step_frequency()
        self.get_step_amplitude()
        self.get_acceleration()
        self.get_velocity()
        self.get_move_mode()
        self.get_holding_time()
        self.command(u":CHAN{}:SDIR?\r\n".format(self.channel), callback=_finish_configuration)

    def start_polling(self):
        super(DeviceSmaractMcs2Channel, self).start_polling()
        self.poll_command(":CHAN{}:STAT?\r\n".format(self.channel), 250)
        self.poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
        self.poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)
        self.poll_command(":CHAN{}:TEMP?\r\n".format(self.channel), 10000)

    def stop_polling(self):
        self.remove_poll_command(":CHAN{}:STAT?\r\n".format(self.channel), 250)
        self.remove_poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
        self.remove_poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)
        self.remove_poll_command(":CHAN{}:TEMP?\r\n".format(self.channel), 10000)

    def subject_update(self, key, value, subject):
        super(DeviceSmaractMcs2Channel, self).subject_update(key, value, subject)

        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]

            if attribute == ATTR_STATE and self.get_value(ATTR_STATE) != value:
                if self.get_value(ATTR_SENSOR_PRESENT) != bool(value & 0x0020):
                    if bool(value & 0x0020):
                        self.poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
                        self.poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)
                    else:
                        self.remove_poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
                        self.remove_poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)

                self.set_raw_value(ATTR_MOVING, bool(value & 0x0001))
                self.set_raw_value(ATTR_CLOSED_LOOP_ACTIVE, bool(value & 0x0002))
                self.set_raw_value(ATTR_CALIBRATING, bool(value & 0x0004))
                self.set_raw_value(ATTR_REFERENCING, bool(value & 0x0008))
                self.set_raw_value(ATTR_MOVE_DELAYED, bool(value & 0x0100))
                self.set_raw_value(ATTR_SENSOR_PRESENT, bool(value & 0x0020))
                self.set_raw_value(ATTR_CALIBRATED, bool(value & 0x0040))
                self.set_raw_value(ATTR_REFERENCED, bool(value & 0x0080))
                self.set_raw_value(ATTR_END_STOP_REACHED, bool(value & 0x0100))
                self.set_raw_value(ATTR_RANGE_LIMIT_REACHED, bool(value & 0x0200))
                self.set_raw_value(ATTR_FOLLOWING_LIMIT_REACHED, bool(value & 0x0400))
                self.set_raw_value(ATTR_MOVEMENT_FAILED, bool(value & 0x0800))
                self.set_raw_value(ATTR_STREAMING, bool(value & 0x1000))
                self.set_raw_value(ATTR_OVER_TEMPERATURE, bool(value & 0x4000))
                self.set_raw_value(ATTR_INDEX_MARK, bool(value & 0x8000))

                holding = bool(value & 0x0004) and bool(value & 0x0008) and bool(value & 0x0008) and bool(
                    value & 0x0002) \
                          and not bool(value & 0x0001)
                self.set_raw_value(ATTR_HOLDING, holding)

                if bool(value & 0x0001):
                    self.set_status(STATUS_BUSY)
                else:
                    self.set_status(STATUS_IDLE)
            elif attribute == ATTR_SENSOR_MODE:
                if value == "Disabled":
                    self.remove_poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
                    self.remove_poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)
                else:
                    self.poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
                    self.poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)
            elif attribute == ATTR_ENCODER_BASE_UNIT:
                if value == 2:
                    value = u"m"
                else:
                    value = u"deg"
            elif attribute == ATTR_ENCODER_RESOLUTION:
                unit = self.get_value(ATTR_ENCODER_BASE_UNIT)
                prefix = {-12: u"p", -9: u"n", -6: u"u", -3: u"m"}
                base_position_unit = prefix.get(value, "") + unit

                self.set_attribute((ATTR_TARGET_POSITION, UNIT), base_position_unit)
                self.set_attribute((ATTR_POSITION_UPPER_LIMIT, UNIT), base_position_unit)
                self.set_attribute((ATTR_POSITION_LOWER_LIMIT, UNIT), base_position_unit)
                self.set_attribute((ATTR_ACCELERATION, UNIT), u"{}/s".format(base_position_unit))
                self.set_attribute((ATTR_VELOCITY, UNIT), u"{}/s".format(base_position_unit))
                self.set_attribute((ATTR_POSITION, UNIT), base_position_unit)
            elif attribute == ATTR_POSITION_LOWER_LIMIT:
                self.set_attribute((ATTR_POSITION, MIN), value)
            elif attribute == ATTR_POSITION_UPPER_LIMIT:
                self.set_attribute((ATTR_POSITION, MAX), value)
            self.set_value(ATTR_LATENCY, self.device.get_value(ATTR_LATENCY))
            self.set_raw_value(attribute, value)

    def enable_close_loop(self):
        if self.position_attribute_copy != None:
            self.set_attribute((ATTR_POSITION, UNIT), self.position_attribute_copy[UNIT])
            self.set_attribute((ATTR_POSITION, OFFSET), self.position_attribute_copy[OFFSET])
            self.set_attribute((ATTR_POSITION, FACTOR), self.position_attribute_copy[FACTOR])
            self.set_value(ATTR_MOVE_MODE, "Absolute")
            self[ATTR_POSITION].set_function = self._move_absolute
            self.position_attribute_copy = None

            # self.poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
            # self.poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)

    def enable_open_loop(self):
        self.position_attribute_copy = self[ATTR_POSITION].attribute_copy()
        default_unit = self.get_value(ATTR_CLOSE_LOOP_UNIT)

        self.set_attribute((ATTR_POSITION, UNIT),
                           default_unit if default_unit != None else self.position_attribute_copy[UNIT])
        self.set_attribute((ATTR_POSITION, OFFSET), 0)
        self.set_attribute((ATTR_POSITION, FACTOR), 1)

        self[ATTR_POSITION].set_function = self._move_steps
        self.set_value(ATTR_MOVE_MODE, "Step")

        # self.remove_poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
        # self.remove_poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)

    def set_move_method(self, value):
        if value == OPEN_LOOP:
            self.enable_open_loop()
        elif value == CLOSE_LOOP:
            self.enable_close_loop()

    @expose_method()
    def stop(self, callback=None):
        self.logger.info(u"Stop movement")
        return self.command(u":STOP{}\r\n".format(self.channel), callback, with_token=True)

    @expose_method()
    def calibrate(self, callback=None):
        self.logger.info(u"Calibrate")
        return self.command(u":CAL{}\r\n".format(self.channel), callback, with_token=True)

    @expose_method()
    def find_reference(self, callback=None):
        self.logger.info(u"Reference")
        return self.command(u":REF{}\r\n".format(self.channel), callback, with_token=True)

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position):
        try:
            move_mode = self[ATTR_MOVE_MODE][VALUE]
            position = device_units(self, ATTR_POSITION, position).m
            self.logger.info(u"Move to absolute position {}".format(position))
            if move_mode == "Relative":
                position = position - self[ATTR_POSITION][VALUE]
            return self.set_value(ATTR_TARGET_POSITION, position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position):
        try:
            move_mode = self[ATTR_MOVE_MODE][VALUE]
            position = device_units(self, ATTR_POSITION, position).m
            self.logger.info(u"Move by relative position {}".format(position))
            if move_mode == "Absolute":
                position = self[ATTR_POSITION][VALUE] + position
            return self.set_value(ATTR_TARGET_POSITION, position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    def _move_absolute(self, value, callback=None):
        self.logger.info(u"Move to position {}".format(value))
        if self.get_value(ATTR_MOVE_MODE) == "Step":
            self.set_value(ATTR_MOVE_MODE, "Absolute")
        token = self.command(u":MOVE{} {}\r\n".format(self.channel, value), callback, with_token=True)
        return token

    def _move_steps(self, value, callback=None):
        if self.get_value(ATTR_MOVE_MODE) != "Step":
            self.set_value(ATTR_MOVE_MODE, "Step")
        current_position = self.get_value(ATTR_POSITION)
        relative_steps = value - current_position
        n_factor, p_factor = self.get_value(ATTR_CLOSE_LOOP_FACTOR)
        relative_steps = relative_steps / (n_factor if relative_steps < 0 else p_factor)
        self.logger.info(f"Move by {relative_steps} steps")
        if not self.get_value(ATTR_SENSOR_PRESENT) or self.get_value(ATTR_SENSOR_MODE) == "Disabled":
            self.set_raw_value(ATTR_POSITION, value)
        token = self.command(u":MOVE{} {}\r\n".format(self.channel, relative_steps), callback, with_token=True)
        return token

    @expose_method({"position": "New position"})
    def set_position(self, position=0):
        self.logger.info(f"Reset position to {position}")
        position = int(position)
        self.command(f":CHAN{self.channel}:POS {position}\r\n")

    def get_encoder_unit(self):
        self.command(u":CHAN{}:TUN:BASE:UNIT?\r\n".format(self.channel))
        self.command(u":CHAN{}:TUN:BASE:RES?\r\n".format(self.channel))

    def get_position_limit_range(self):
        self.command(u":CHAN{}:RLIM:MAX?\r\n".format(self.channel))
        self.command(u":CHAN{}:RLIM:MIN?\r\n".format(self.channel))

    def get_positioner_configuration(self):
        self.command(u":CHAN{}:PTYP:NAME?\r\n".format(self.channel))
        self.command(u":CHAN{}:PTYP:CODE?\r\n".format(self.channel))

    def get_step_frequency(self):
        self.command(u":CHAN{}:STEP:FREQ?\r\n".format(self.channel))

    def get_step_amplitude(self):
        self.command(u":CHAN{}:STEP:AMPL?\r\n".format(self.channel))

    def get_velocity(self):
        self.command(u":CHAN{}:VEL?\r\n".format(self.channel))

    def get_acceleration(self):
        self.command(u":CHAN{}:ACC?\r\n".format(self.channel))

    def get_sensor_mode(self):
        self.command(u":CHAN{}:SENS:MODE?\r\n".format(self.channel))

    def get_move_mode(self):
        self.command(u":CHAN{}:MMOD?\r\n".format(self.channel))

    def get_safe_direction(self):
        self.command(u":CHAN{}:SDIR?\r\n".format(self.channel))

    def get_holding_time(self):
        self.command(u":CHAN{}:HOLD?\r\n".format(self.channel))

    def set_holding_time(self, value, callback=None):
        self.logger.info(u"Set holding time to {}".format(value))
        return self.command(u":CHAN{}:HOLD {}\r\n".format(self.channel, int(value)), callback, with_token=True)

    def set_upper_position_limit(self, value, callback=None):
        self.logger.info(u"Set upper limit to {}".format(value))
        token = self.command(u":CHAN{}:RLIM:MAX {}\r\n".format(self.channel, int(value)), callback, with_token=True)
        self.set_attribute([ATTR_POSITION, MAX], value)
        return token

    def set_lower_position_limit(self, value, callback=None):
        self.logger.info(u"Set lower limit to {}".format(value))
        token = self.command(u":CHAN{}:RLIM:MIN {}\r\n".format(self.channel, int(value)), callback, with_token=True)
        self.set_attribute([ATTR_POSITION, MIN], value)
        return token

    def set_positioner_code(self, value, callback=None):
        self.logger.info(u"Set positioner  code to {}".format(value))
        token = self.command(u":CHAN{}:PTYP:CODE {}\r\n".format(self.channel, int(value)), callback, with_token=True)
        self.get_positioner_configuration()
        self.get_encoder_unit()
        return token

    def set_step_frequency(self, value, callback=None):
        self.logger.info(u"Set step frequency to {} Hz".format(value))
        self.command(u":CHAN{}:MCLF:CURR {}\r\n".format(self.channel, int(value)), callback, with_token=True)
        return self.command(u":CHAN{}:STEP:FREQ {}\r\n".format(self.channel, int(value)), callback, with_token=True)

    def set_step_amplitude(self, value, callback=None):
        self.logger.info(u"Set step amplitude to {}".format(value))
        return self.command(u":CHAN{}:STEP:AMPL {}\r\n".format(self.channel, int(value)), callback, with_token=True)

    def set_velocity(self, value, callback=None):
        self.logger.info(u"Set velocity to {}".format(value))
        return self.command(u":CHAN{}:VEL {}\r\n".format(self.channel, int(value)), callback, with_token=True)

    def set_acceleration(self, value, callback=None):
        self.logger.info(u"Set acceleration to {}".format(value))
        return self.command(u":CHAN{}:ACC {}\r\n".format(self.channel, int(value)), callback, with_token=True)

    def set_sensor_mode(self, value, callback=None):
        self.logger.info(u"Set sensor mode to {}".format(value))
        value = self.sensor_modes[value]
        out = self.command(u":CHAN{}:SENS:MODE {}\r\n".format(self.channel, value), callback, with_token=True)
        self.get_sensor_mode()
        return out

    def set_move_mode(self, value, callback=None):
        self.logger.info(u"Set move mode to {}".format(value))
        if value == "Step":
            self.set_value(ATTR_MOVE_METHOD, OPEN_LOOP)
        else:
            self.set_value(ATTR_MOVE_METHOD, CLOSE_LOOP)

        value = self.move_modes[value]
        return self.command(u":CHAN{}:MMOD {}\r\n".format(self.channel, value), callback, with_token=True)

    def set_safe_direction(self, value, callback=None):
        self.logger.info(u"Set safe direction to {}".format(value))
        value = self.safe_direction_modes[value]
        return self.command(u":CHAN{}:SDIR {}\r\n".format(self.channel, value), callback, with_token=True)

    @expose_method()
    def enable_holding(self, callback=None):
        """
        #
        :param callback:
        :return:
        """
        # enable holding
        self.set_value(ATTR_HOLDING_TIME, -1)
        # move by 0 to end holding
        self.set_value(ATTR_MOVE_MODE, "Relative")
        self.move_relative(device_units(self, ATTR_POSITION, 0))
        self.set_value(ATTR_MOVE_MODE, "Absolute")

    @expose_method()
    def disable_holding(self, callback=None):
        """
        #
        :param callback:
        :return:
        """
        # disable holding
        self.set_value(ATTR_HOLDING_TIME, 0)
        # move by 0 to end holding
        self.set_value(ATTR_MOVE_MODE, "Relative")
        self.move_relative(device_units(self, ATTR_POSITION, 0))
        self.set_value(ATTR_MOVE_MODE, "Absolute")


class DeviceMCS2FlyScanner(DeviceScanner):

    def __init__(self, device_id=None, config=None):
        super().__init__(device_id, config)

    def _init_attributes(self):
        DeviceScanner._init_attributes(self)
        last_scan = kamzik3.session.get_scan_count()
        self.create_attribute(ATTR_DWELL_TIME, default_value=100, default_type=np.float32, min_value=1, unit="ms")

    @expose_method()
    def get_scanner_attributes(self):
        raise NotImplementedError

    @expose_method()
    def get_scanner_macro(self, scanner_input, scanner_attributes, parent_macro=None):
        raise NotImplementedError



from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.observer import Observer


class DeviceChannel(Device, Observer):

    def __init__(self, device, channel, device_id=None, config=None):
        self.configured = False
        self.device = device
        self.channel = channel
        self.position_attribute_copy = None
        super(DeviceChannel, self).__init__(device_id, config)
        self.set_status(STATUS_CONNECTING)
        self.device.attach_observer(self)

    def _init_attributes(self):
        Device._init_attributes(self)
        # self.create_attribute(ATTR_STEPS, default_value=0, default_type=np.int64, set_function=self._move_absolute)
        # self.create_attribute(ATTR_CLOSE_LOOP_UNIT, display=False)
        # self.create_attribute(ATTR_CLOSE_LOOP_FACTOR, default_value=[1, 1], default_type=TYPE_LIST, display=False)
        # self.create_attribute(ATTR_MOVE_METHOD, default_value=CLOSE_LOOP, default_type=[OPEN_LOOP, CLOSE_LOOP],
        #                       set_function=self.set_move_method)

    # def set_move_method(self, value):
    #     if value == OPEN_LOOP:
    #         self.enable_open_loop()
    #     elif value == CLOSE_LOOP:
    #         self.enable_close_loop()
    #
    # def enable_close_loop(self):
    #     if self.position_attribute_copy != None:
    #         self.set_attribute((ATTR_POSITION, UNIT), self.position_attribute_copy[UNIT])
    #         self.set_attribute((ATTR_POSITION, OFFSET), self.position_attribute_copy[OFFSET])
    #         self.set_attribute((ATTR_POSITION, FACTOR), self.position_attribute_copy[FACTOR])
    #         self.position_attribute_copy = None
    #
    # def enable_open_loop(self):
    #     self.position_attribute_copy = self[ATTR_POSITION].attribute_copy()
    #     default_unit = self.get_value(ATTR_CLOSE_LOOP_UNIT)
    #     self.set_attribute((ATTR_POSITION, UNIT), default_unit if default_unit != None else self.position_attribute_copy[UNIT])
    #     self.set_attribute((ATTR_POSITION, OFFSET), 0)
    #     self.set_attribute((ATTR_POSITION, FACTOR), 1)

    # def _move_absolute(self):
    #     raise NotImplementedError

    def command(self, command, callback=None, with_token=False, returning=True):
        return self.device.command(command, callback, with_token, returning)

    def subject_update(self, key, value, subject):
        if key == ATTR_STATUS:
            if value in READY_DEVICE_STATUSES:
                self.handle_configuration()
            else:
                self.configured = False
                self.set_status(value)

    def poll_command(self, command, interval):
        self.device.poll_command(command, interval)

    def remove_poll_command(self, command, interval):
        self.device.remove_poll_command(command, interval)

    def handle_configuration(self):
        raise NotImplementedError(u"Must be implemented in subclass")

    def disconnect(self):
        self.stop_polling()
        self.configured = False
        self.device.detach_observer(self)
        self.set_status(STATUS_DISCONNECTED)

    def reconnect(self, *args):
        self.device.attach_observer(self)
        self.handle_configuration()



import os

import pandas as pd


def convert_yml_to_hdf5(path_to_yml, path_to_hdf5):
    """
    Convert old YAML formatted save file into new HDF5 format.
    @param path_to_yml: Path to the input YAML file
    @type path_to_yml: str
    @param path_to_hdf5: Path to the output HDF5 file
    @type path_to_hdf5: str
    """
    with open(path_to_yml, "r") as input_file:
        input_attributes = yaml.load(input_file, Loader=yaml.Loader)

    groups = input_attributes["Preset groups"]
    subgroup_units = {}
    for group, subgroups in groups.items():
        for subgroup, values in subgroups.items():
            group_path = os.path.join(group, subgroup)
            columns = values["columns"]
            rows = values["rows"]

            new_columns = {"Color": ["None"], "Time": ["None"], "Comment": ["None"]}
            for device_id, attribute in columns[1:]:
                device_attribute = f"{device_id}{attribute[:-1]}"
                if device_attribute not in subgroup_units:
                    unit = input(f"Set unit for {device_attribute} (mm): ")
                    if unit == "":
                        unit = "mm"
                    subgroup_units[device_attribute] = unit
                new_columns[f"{device_id} {attribute[:-1]}[{subgroup_units[device_attribute]}]"] = ["None"]
            for row in rows:
                for column_index, column_id in enumerate(new_columns.keys()):
                    row_value = row[column_index]
                    if column_index > 2 and row_value not in (None, "None", ""):
                        try:
                            row_value, _ = row_value.split(" ")
                        except ValueError:
                            print(repr(row_value))
                    new_columns[column_id].append(str(row_value))

            df = pd.DataFrame(new_columns, columns=list(new_columns.keys()))
            # Save group into HDF5 file.
            # Use format table and compression level 1
            # I was testing all compression levels without much gain of size.
            # Level 1 however gives almost 30% size decrease
            df.to_hdf(path_to_hdf5, group_path, format="table", complevel=1)


def generate_saved_attributes(path, groups):
    """
    Add new group into saved attributes file.
    Example of groups:
    {"/Default/Sample": {"X ['Position'][mm]", "Y ['Position'][mm]", "Yaw ['Position'][deg]"}}
    Note that Attribute name has to be in single parentheses.
    Example of nested attribute can be "X ['Motor group', 'Position']"
    @param path: path to the saved attribute file
    @param groups: dict containing group : header pairs
    """
    for group, data in groups.items():
        df = pd.DataFrame(data, columns=list(data.keys()))
        df.to_hdf(path, group, format="table")


if __name__ == "__main__":
    import argparse

    import oyaml as yaml

    # Parse arguments from commandline
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", help="Input path for saved attributes",
                        default="./saved_attributes.att")
    parser.add_argument("--output", help="Path where to save new modified file",
                        default="./saved_attributes_mod.att")
    parser.add_argument("--chdir", help="Path to active directory",
                        default="./")
    args = parser.parse_args()

    # Set active directory
    os.chdir(args.chdir)

    generate_saved_attributes(
        "/usr/local/share/pycharm-projects/bt_p11_june_2021/bt_p11_june_2021/saved_attributes.h5", {
            "/Default/Apertures": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                   "APT-X ['Position'][mm]": ["None"],
                                   "APT-Y ['Position'][mm]": ["None"],
                                   "OSB-X ['Position'][mm]": ["None"],
                                   "OSB-Y ['Position'][mm]": ["None"],
                                   "OSB-Z ['Position'][mm]": ["None"],
                                   },
            "/Default/Phase Plate": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                     "PP-X ['Position'][mm]": ["None"],
                                     "PP-Y ['Position'][mm]": ["None"],
                                     "PP-Z ['Position'][mm]": ["None"],
                                     "PP-Roll ['Position'][deg]": ["None"],
                                     "PP-Pitch ['Position'][deg]": ["None"],
                                     "PP-Yaw ['Position'][deg]": ["None"]
                                     },
            "/Default/MLL Down": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                  "MLL-DOWN-X ['Position'][mm]": ["None"],
                                  "MLL-DOWN-Y ['Position'][mm]": ["None"],
                                  "MLL-DOWN-Z ['Position'][mm]": ["None"],
                                  "MLL-DOWN-Roll ['Position'][deg]": ["None"],
                                  "MLL-DOWN-Pitch ['Position'][deg]": ["None"],
                                  "MLL-DOWN-Yaw ['Position'][deg]": ["None"]
                                  },
            "/Default/MLL Up": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                "MLL-UP-X ['Position'][mm]": ["None"],
                                "MLL-UP-Y ['Position'][mm]": ["None"],
                                "MLL-UP-Z ['Position'][mm]": ["None"],
                                "MLL-UP-Roll ['Position'][deg]": ["None"],
                                "MLL-UP-Pitch ['Position'][deg]": ["None"],
                                "MLL-UP-Yaw ['Position'][deg]": ["None"]
                                },
            "/Default/Ifer Down": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                        "RD-Y ['Position'][mm]": ["None"],
                                        "RD-Z ['Position'][mm]": ["None"],
                                        "MD-Z ['Position'][mm]": ["None"],
                                        },
            "/Default/Ifer Up": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                      "RU-X ['Position'][mm]": ["None"],
                                      "RU-Z ['Position'][mm]": ["None"],
                                      "MU-Z ['Position'][mm]": ["None"],
                                      },
            "/Default/Sample": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                "SAM-X ['Position'][mm]": ["None"],
                                "SAM-Y ['Position'][mm]": ["None"],
                                "SAM-Z ['Position'][mm]": ["None"],
                                "SAM-R ['Position'][deg]": ["None"],
                                },
            "/Default/Cone": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                              "CONE-X ['Position'][mm]": ["None"],
                              "CONE-Y ['Position'][mm]": ["None"],
                              "CONE-Z ['Position'][mm]": ["None"],
                              },
            "/Default/Microscope": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                    "MIC-X ['Position'][mm]": ["None"],
                                    "MIC-Y ['Position'][mm]": ["None"],
                                    "MIC-Z ['Position'][mm]": ["None"],
                                    },
            "/Default/Beamstop": {"Color": ["None"], "Time": ["None"], "Comment": ["None"],
                                  "BS-X ['Position'][mm]": ['None'],
                                  "BS-Y ['Position'][mm]": ["None"]},
        })
    # convert_yml_to_hdf5(args.input, args.output)



import os
from datetime import datetime
from pathlib import PurePath
import re
import natsort
from PyQt5.QtCore import QAbstractItemModel, QAbstractListModel, Qt, QAbstractTableModel
from PyQt5.QtWidgets import QDialog, QHeaderView, QFileIconProvider

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.gui.templates.customOpenDialogTemplate import Ui_Dialog
from kamzik3.snippets.snippetsUnits import get_printable_size
from kamzik3.snippets.snippetsWidgets import show_error_message


class CategoryModel(QAbstractListModel):

    def __init__(self, model_data):
        self.model_data = list(model_data)
        self.icon_provider = QFileIconProvider()
        QAbstractItemModel.__init__(self)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.model_data)

    def data(self, model_index, role=None):
        if role == Qt.DisplayRole:
            return self.model_data[model_index.row()]
        elif role == Qt.DecorationRole:
            return self.icon_provider.icon(QFileIconProvider.Drive)


class DirContentModel(QAbstractTableModel):

    def __init__(self, model_data):
        self.sorted_by = (0, 1)
        self.model_data_unfiltered = list(model_data)
        self.model_data = list(model_data)
        self.columns = ["Name", "Size", "Modified"]
        self.icon_provider = QFileIconProvider()
        self.filter_regexp = None
        QAbstractTableModel.__init__(self)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.model_data[1]) + len(self.model_data[3])

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.columns)

    def headerData(self, p_int, Qt_Orientation, role=None):
        if role == Qt.DisplayRole and Qt_Orientation == Qt.Horizontal:
            return self.columns[p_int]

    def data(self, index, role):
        if role == Qt.DisplayRole:
            name, (file_size, mod_time) = self.get_item(index)
            if index.column() == 0:
                return name
            elif index.column() == 1:
                return get_printable_size(file_size)
            elif index.column() == 2:
                dt_object = datetime.fromtimestamp(mod_time)
                return str(dt_object.strftime('%d.%m.%Y %H:%M'))
        elif role == Qt.DecorationRole and index.column() == 0:
            if self.is_folder(index):
                return self.icon_provider.icon(QFileIconProvider.Folder)
            else:
                return self.icon_provider.icon(QFileIconProvider.File)

    def is_folder(self, model_index):
        return model_index.row() < len(self.model_data[1])

    def get_item(self, model_index):
        dir_offset = len(self.model_data[1])
        file_index = model_index.row() - dir_offset
        if file_index < 0:
            return self.model_data[1][model_index.row()], self.model_data[2][model_index.row()]
        else:
            return self.model_data[3][file_index], self.model_data[4][file_index]

    def flags(self, index):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def sort(self, p_int, order=None):
        if p_int == 0:
            if len(self.model_data[1]) > 0:
                self.model_data[1], self.model_data[2] = zip(
                    *natsort.humansorted(zip(self.model_data[1], self.model_data[2]), reverse=not order))
            if len(self.model_data[3]) > 0:
                self.model_data[3], self.model_data[4] = zip(
                    *natsort.humansorted(zip(self.model_data[3], self.model_data[4]), reverse=not order))
        if p_int == 1:
            if len(self.model_data[3]) > 0:
                self.model_data[3], self.model_data[4] = zip(
                    *sorted(zip(self.model_data[3], self.model_data[4]), reverse=not order, key=lambda x: x[1][0]))
        if p_int == 2:
            if len(self.model_data[1]) > 0:
                self.model_data[1], self.model_data[2] = zip(
                    *sorted(zip(self.model_data[1], self.model_data[2]), reverse=not order, key=lambda x: x[1][1]))
            if len(self.model_data[3]) > 0:
                self.model_data[3], self.model_data[4] = zip(
                    *sorted(zip(self.model_data[3], self.model_data[4]), reverse=not order, key=lambda x: x[1][1]))
        self.sorted_by = (p_int, order)
        self.layoutChanged.emit()

    def name_filter(self, name):
        self.filter_regexp = re.compile(f".*{name}.*")
        self.model_data[3], self.model_data[4] = [], []
        for index, file_name in enumerate(self.model_data_unfiltered[3]):
            if re.search(self.filter_regexp, file_name) is not None:
                self.model_data[3].append(self.model_data_unfiltered[3][index])
                self.model_data[4].append(self.model_data_unfiltered[4][index])
        self.sort(*self.sorted_by)


class CustomOpenDialog(QDialog, Ui_Dialog):
    current_path = None
    current_drive = None
    selected_files = None

    def __init__(self, virtual_directory_device, title="Select file", parent=None):
        self.virtual_directory_device = virtual_directory_device
        QDialog.__init__(self, parent)
        self.setupUi(self)
        self.setWindowTitle(title)

        drives = self.virtual_directory_device.get_value(ATTR_DRIVES)
        category_model = CategoryModel(list(drives.keys()))
        self.input_category.setModel(category_model)
        model_index = category_model.createIndex(0, 0)
        self.input_category.setCurrentIndex(model_index)
        self.slot_drive_changed(model_index)

    def showEvent(self, event):
        button_open = self.buttonBox.button(self.buttonBox.Open)
        button_open.clearFocus()
        button_open.setAutoDefault(False)
        button_open.setDefault(False)
        self.input_file_name.setFocus()

    def slot_file_filter(self):
        table_model = self.table_directory_content.model()
        table_model.name_filter(self.input_file_name.text())

    def slot_drive_changed(self, model_index):
        self.current_drive = model_index.data(Qt.DisplayRole)
        drives = self.virtual_directory_device.get_value(ATTR_DRIVES)
        self.slot_change_path(drives[self.current_drive]["path"])

    def slot_file_selected(self, model_index):
        table_model = self.table_directory_content.model()
        selected_index = table_model.createIndex(model_index.row(), 0)
        filename = table_model.data(selected_index, Qt.DisplayRole)
        if table_model.is_folder(selected_index):
            self.slot_change_path(os.path.join(self.current_path, filename))
        else:
            self.accept()

    def slot_change_path(self, path):
        try:
            path = os.path.normpath(path)
            dir_tree = self.virtual_directory_device.get_dir_content(drive=self.current_drive, input_dir=path)
            self.current_path = path
            self.set_dir_content(dir_tree)
        except DeviceError:
            show_error_message("Virtual directory device is not ready.", parent=self)

    def slot_dir_up(self):
        new_root_dir = os.path.split(self.current_path)[0]
        self.slot_change_path(new_root_dir)

    def set_dir_content(self, dir_tree):
        current_model = self.table_directory_content.model()
        sorted_by = (0, 1)
        if current_model is not None:
            sorted_by = current_model.sorted_by
        dir_content_model_data = dir_tree
        self.table_directory_content.setModel(DirContentModel(dir_content_model_data))
        self.table_directory_content.model().sort(*sorted_by)
        header = self.table_directory_content.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.Stretch)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QHeaderView.ResizeToContents)
        self.table_directory_content.selectRow(0)
        self.set_navigation()

    def set_navigation(self):
        category_root_path = self.virtual_directory_device.get_value(ATTR_DRIVES)[self.current_drive]["path"]
        self.button_up.setDisabled(self.current_path == category_root_path)
        category_root_parts = PurePath(self.current_path.replace(category_root_path, "")).parts
        self.input_directory.blockSignals(True)
        self.input_directory.clear()
        paths = [category_root_path]
        for part in category_root_parts[1:]:
            paths.insert(0, os.path.join(paths[0], part))
        self.input_directory.addItems(paths)
        self.input_directory.blockSignals(False)

    def accept(self):
        selected_index = self.table_directory_content.selectedIndexes()[0]
        table_model = self.table_directory_content.model()
        selected_index = table_model.createIndex(selected_index.row(), 0)
        filename = table_model.data(selected_index, Qt.DisplayRole)
        file_path = os.path.join(self.current_path, filename)
        if table_model.is_folder(selected_index):
            self.set_dir_content(self.dir_indexes.get(file_path, 0))
        else:
            self.selected_files = [(self.current_drive, file_path)]
            QDialog.accept(self)



import time
from threading import Thread, Lock
from time import sleep

import bidict
import numpy as np
import smaract.picoscale as ps
import smaract.si as si

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer

use_data_source_types = {
    si.DataSource.POSITION: "Position",
    si.DataSource.TEMPERATURE: "Temperature",
    si.DataSource.HUMIDITY: "Humidity",
    si.DataSource.PRESSURE: "Pressure",
    # si.DataSource.COS_QUALITY: "Cos quality",
    si.DataSource.SIN_QUALITY: "Sin quality"
}


datatypes = {
    si.DataType.INT8: np.int8,
    si.DataType.UINT8: np.uint8,
    si.DataType.INT16: np.int16,
    si.DataType.UINT16: np.uint16,
    si.DataType.INT32: np.int32,
    si.DataType.UINT32: np.uint32,
    si.DataType.INT48: np.int64,
    si.DataType.UINT48: np.uint64,
    si.DataType.INT64: np.int64,
    si.DataType.UINT64: np.uint64,
    si.DataType.FLOAT32: np.float32,
    si.DataType.FLOAT64: np.float64,
    si.DataType.STRING: str,
}

si_units = {
    si.BaseUnit.METRE: "m",
    si.BaseUnit.NO: None,
    si.BaseUnit.DEGREE: "deg",
    si.BaseUnit.DEGREE_CELSIUS: "degC",
    si.BaseUnit.AMPERE: "A",
    si.BaseUnit.HERTZ: "Hz",
    si.BaseUnit.KELVIN: "k",
    si.BaseUnit.KILOGRAM: "Kg",
    si.BaseUnit.PERCENT: "%",
    si.BaseUnit.SECOND: "s",
    si.BaseUnit.WATT: "W",
    si.BaseUnit.VOLT: "V",
    si.BaseUnit.PASCAL: "Pa",
}

streaming_modes = bidict.FrozenOrderedBidict({
    si.StreamingMode.DIRECT: "Direct",
    si.StreamingMode.TRIGGERED: "Triggered",
})

GROUP_STREAMING = "Streaming"


class DevicePicoscale(Device):
    # Increase connection timeout, since this Device takes it's time.
    # Usually it takes about 20 seconds.
    # To be safe, set connection timeout to 25 seconds
    connection_timeout = 25000
    polling = False
    polling_timer = None

    def __init__(self, locator, device_id=None, config=None):
        """
        Locator defines where to find Picoscale unit.
        From documentation some examples:
            usb:ix:0
            network:192.168.1.100:55555
        @param locator: locator address
        @type locator: str
        @param device_id: device ID
        @type device_id: str
        @param config: config dict
        @type config: dict
        """
        self.locator = locator
        self.handle = None
        self.polled_functions = []
        self.stream_sources = []
        self.stream_buffers = {}
        self.available_head_types = bidict.OrderedBidict()
        self.readout_lock = Lock()
        Device.__init__(self, device_id, config)
        self.connect()

    def connect(self, *args):
        """
        Use si.open to connect to PicoScale.
        It returns handle.
        According to what we observer, only one client can be connected at the same time.
        It's also taking a long time to connect. At some cases close to 20 seconds.
        :param args: connect attributes
        """
        try:
            self.connecting = True
            self.connected = False
            self.device_connection_poller.add_connecting_device(self)
            self.handle = si.Open(self.locator)
            self.handle_connect_event()
        except si.Error as e:
            self.logger.exception(f"Connection exception: {self._get_exception(e)}")
        except DeviceError:
            self.logger.exception(u"Connection exception")
            return

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_CHANNELS, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_FULL_ACCESS, default_value=True, default_type=bool,
                              set_function=self.set_full_access)
        self.create_attribute(ATTR_PILOT_LASER, default_value=False, default_type=bool,
                              set_function=self.set_pilot_laser)
        self.create_attribute(ATTR_IS_STABLE, readonly=True, default_type=bool)
        self.create_attribute(ATTR_FRAME_AGGREGATION, group=GROUP_STREAMING, default_value=1, default_type=np.uint32,
                              set_function=self.set_frame_aggregation, min_value=1, max_value=1)
        self.create_attribute(ATTR_FRAMERATE, group=GROUP_STREAMING, default_value=1, default_type=np.uint32,
                              set_function=self.set_frame_rate, unit="Hz", min_value=1, max_value=10e6)
        self.create_attribute(ATTR_BUFFER_AGGREGATION, group=GROUP_STREAMING, default_value=0, default_type=np.uint32,
                              set_function=self.set_buffer_aggregation, min_value=0)
        self.create_attribute(ATTR_STREAMING_MODE, group=GROUP_STREAMING,
                              default_value=streaming_modes[si.StreamingMode.DIRECT],
                              default_type=list(streaming_modes.values()), set_function=self.set_streaming_mode)
        self.create_attribute(ATTR_BUFFERS_INTERLEAVED, group=GROUP_STREAMING, default_value=False, default_type=bool,
                              set_function=self.set_buffers_interleaved)
        self.create_attribute(ATTR_FILTER_RATE, group=GROUP_STREAMING, default_value=10e6, default_type=np.float64,
                              set_function=self.set_filter_rate, min_value=1, max_value=10e6, unit="Hz")
        self.create_attribute(ATTR_BUFFERS_COUNT, group=GROUP_STREAMING, default_value=2, default_type=np.uint32,
                              set_function=self.set_buffers_count, min_value=2, max_value=255)
        self.create_attribute(ATTR_HEAD_FIBER_LENGTH, default_value=1660, default_type=np.float64,
                              set_function=self.set_head_fiber_length, min_value=0, unit="mm", decimals=1)
        self.create_attribute(ATTR_EXTENSION_FIBER_LENGTH, default_value=1660, default_type=np.float64,
                              set_function=self.set_extension_fiber_length, min_value=0, unit="mm", decimals=1)
        self.create_attribute(ATTR_MIN_WORKING_DISTANCE, default_value=1, default_type=np.float64,
                              set_function=self.set_min_working_distance, min_value=1, max_value=500, unit="mm",
                              decimals=1)
        self.create_attribute(ATTR_MAX_WORKING_DISTANCE, default_value=1, default_type=np.float64,
                              set_function=self.set_max_working_distance, min_value=1, max_value=500, unit="mm",
                              decimals=1)

    def handle_configuration(self):
        start_at = time.time()
        self._config_commands()
        self._config_attributes()
        try:
            self.set_full_access(True)
        except DeviceError:
            self.logger.error("Full access could not be acquired!")
        serial_number = si.GetProperty_s(self.handle, si.EPK(si.Property.DEVICE_SERIAL_NUMBER, 0, 0))
        self.set_value(ATTR_SERIAL_NUMBER, serial_number)
        number_of_channels = si.GetProperty_i32(self.handle, si.EPK(si.Property.NUMBER_OF_CHANNELS, 0, 0))
        self.set_value(ATTR_CHANNELS, number_of_channels)
        head_fiber_length = si.GetProperty_f64(self.handle, si.EPK(ps.Property.SYS_FIBERLENGTH_HEAD, 0, 0))
        self.set_raw_value(ATTR_HEAD_FIBER_LENGTH, head_fiber_length)
        extension_fiber_length = si.GetProperty_f64(self.handle, si.EPK(ps.Property.SYS_FIBERLENGTH_EXTENSION, 0, 0))
        self.set_raw_value(ATTR_EXTENSION_FIBER_LENGTH, extension_fiber_length)
        min_working_distance = si.GetProperty_f64(self.handle, si.EPK(ps.Property.SYS_WORKING_DISTANCE_MIN, 0, 0))
        self.set_raw_value(ATTR_MIN_WORKING_DISTANCE, min_working_distance)
        max_working_distance = si.GetProperty_f64(self.handle, si.EPK(ps.Property.SYS_WORKING_DISTANCE_MAX, 0, 0))
        self.set_raw_value(ATTR_MAX_WORKING_DISTANCE, max_working_distance)

        head_type_category_count = si.GetProperty_i32(self.handle,
                                                      si.EPK(ps.Property.SYS_HEAD_TYPE_CATEGORY_COUNT, 0, 0))
        for category_index in range(head_type_category_count):
            head_type_count = si.GetProperty_i32(self.handle,
                                                 si.EPK(ps.Property.SYS_HEAD_TYPE_COUNT, category_index, 0))
            head_category_name = si.GetProperty_s(self.handle,
                                                  si.EPK(ps.Property.SYS_HEAD_TYPE_CATEGORY_NAME, category_index, 0))
            for head_type_index in range(head_type_count):
                head_type_name = si.GetProperty_s(self.handle,
                                                  si.EPK(ps.Property.SYS_HEAD_TYPE_NAME, category_index,
                                                         head_type_index))
                self.available_head_types[head_type_name] = (category_index, head_type_index)
        self.get_pilot_laser()
        self.get_system_stable()
        self.start_polling()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def set_min_working_distance(self, value):
        try:
            si.SetProperty_f64(self.handle, si.EPK(ps.Property.SYS_WORKING_DISTANCE_MIN, 0, 0), value)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SYS_WORKING_DISTANCE_ACTIVATE, 0, 0), 1)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_max_working_distance(self, value):
        try:
            si.SetProperty_f64(self.handle, si.EPK(ps.Property.SYS_WORKING_DISTANCE_MAX, 0, 0), value)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SYS_WORKING_DISTANCE_ACTIVATE, 0, 0), 1)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_head_fiber_length(self, value):
        try:
            si.SetProperty_f64(self.handle, si.EPK(ps.Property.SYS_FIBERLENGTH_HEAD, 0, 0), value)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_extension_fiber_length(self, value):
        try:
            si.SetProperty_f64(self.handle, si.EPK(ps.Property.SYS_FIBERLENGTH_EXTENSION, 0, 0), value)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_buffers_count(self, value: int):
        """
        Set number of buffers we want to use for streaming.
        Usually we want to use maximum number 255.
        :param value: Number of buffers
        :return:
        """
        try:
            si.SetProperty_i32(self.handle, si.EPK(si.Property.NUMBER_OF_STREAMBUFFERS, 0, 0), int(value))
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_buffers_interleaved(self, value: bool):
        """
        Set to True, when You want to interleave multiple sources into one buffer.
        :param value: flag
        :return:
        """
        try:
            si.SetProperty_i32(self.handle, si.EPK(si.Property.STREAMBUFFERS_INTERLEAVED, 0, 0), int(value))
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_filter_rate(self, value: float):
        """
        Set filtering rate self.configure_digital_outputs()of data.
        Usually we want to set same value as framerate.
        :param value: Filteself.configure_digital_outputs()r rate value
        :return:
        """
        try:
            si.SetProperty_f64(self.handle, si.EPK(ps.Property.SYS_FILTER_RATE, 0, 0), value)
            real_filter_rate = si.GetProperty_f64(self.handle, si.EPK(ps.Property.SYS_FILTER_RATE, 0, 0))
            self.set_raw_value([GROUP_STREAMING, ATTR_FILTER_RATE], real_filter_rate)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_frame_aggregation(self, value):
        try:
            si.SetProperty_i32(self.handle, si.EPK(si.Property.FRAME_AGGREGATION, 0, 0), int(value))
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_buffer_aggregation(self, value):
        try:
            si.SetProperty_i32(self.handle, si.EPK(si.Property.STREAMBUFFER_AGGREGATION, 0, 0), int(value))
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_frame_rate(self, value: int) -> None:
        try:
            si.SetProperty_i32(self.handle, si.EPK(si.Property.FRAME_RATE, 0, 0), int(value))
            real_frame_rate = si.GetProperty_f64(self.handle, si.EPK(si.Property.PRECISE_FRAME_RATE, 0, 0))
            self.set_raw_value([GROUP_STREAMING, ATTR_FRAMERATE], real_frame_rate)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_streaming_mode(self, value: str) -> None:
        value = streaming_modes.inv[value]
        try:
            si.SetProperty_i32(self.handle, si.EPK(si.Property.STREAMING_MODE, 0, 0), value)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method({"slot": "configuration slot"})
    def load_configuration_slot(self, slot):
        try:
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SYS_CONFIGURATION_LOAD, int(slot), 0), 1)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method()
    def start_manual_adjustment(self):
        try:
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.AF_ADJUSTMENT_STATE, 0, 0),
                               ps.AdjustmentState.MANUAL_ADJUST)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method()
    def start_auto_adjustment(self):
        try:
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.AF_ADJUSTMENT_STATE, 0, 0),
                               ps.AdjustmentState.AUTO_ADJUST)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    # @expose_method()
    # def start_quality_measurement(self):
    #     # self.save_adjustment_results()
    #     self.start_manual_adjustment()
    #
    # @expose_method()
    # def stop_quality_measurement(self):
    #     self.stop_adjustment()
    #     self.load_adjustment_results()
    #
    #     # def later():
    #     #     sleep(10)
    #     #     for channel_id in range(self.get_value(ATTR_CHANNELS)):
    #     #         try:
    #     #             si.SetProperty_i32(self.handle, si.EPK(ps.Property.AF_CHANNEL_VALIDATION_STATE, channel_id, 0), 1)
    #     #         except si.Error as e:
    #     #             raise DeviceError(self._get_exception(e))
    #     # Thread(target=later).start()

    @expose_method()
    def stop_adjustment(self):
        try:
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.AF_ADJUSTMENT_STATE, 0, 0), ps.AdjustmentState.DISABLED)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method({"clear_values": "Clear previous streamed values"})
    def start_streaming(self, clear_values: bool = True) -> None:
        try:
            clear_values = bool(int(clear_values))
            self.stream_sources = sorted(self.stream_sources, key=lambda element: (element[0], element[1]))
            for channel, source in self.stream_sources:
                if clear_values:
                    self.notify((channel, source, ATTR_STREAM_READOUT), None)
                self.stream_buffers[(channel, source)] = []
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SG_CLOCK_SOURCE, 0, 0),
                               ps.StreamGeneratorSource.INTERNAL)
            si.SetProperty_i32(self.handle, si.EPK(si.Property.STREAMING_MODE, 0, 0), si.StreamingMode.DIRECT)
            si.SetProperty_i32(self.handle, si.EPK(si.Property.STREAMING_ACTIVE, 0, 0), si.ENABLED)
            self.set_status(STATUS_BUSY)
            Thread(target=self.stream_data_thread).start()
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method()
    def clear_streams(self) -> None:
        for channel, source in self.stream_sources:
            self.notify((channel, source, ATTR_STREAM_READOUT), None)
            self.stream_buffers[(channel, source)] = []

    @expose_method()
    def setup_triggered_stream(self) -> None:
        """
        Prepare triggered stream, that expects HW TTL signal _|¯¯¯¯|_.
        Setup inputs and outputs to collect streamed data for width of TTL pulse.
        :return: None
        """
        try:
            self.logger.info("Preparing triggering stream")
            # Setup source of Trigger 0 and 1 to External Trigger
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_SOURCE_EVENT, 0, 0),
                               ps.TriggerEvent.EXTERNAL_TRIGGER)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_SOURCE_EVENT, 1, 0),
                               ps.TriggerEvent.EXTERNAL_TRIGGER)

            # Setup Trigger output to Immediate and don't reset stream at next trigger for trigger output 0 and 1
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_OUTPUT_MODE, 0, 0),
                               ps.TriggerOutputMode.IMMEDIATE_NO_RESET)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_OUTPUT_MODE, 1, 0),
                               ps.TriggerOutputMode.IMMEDIATE_NO_RESET)

            # Setup Trigger start index to 0
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SG_TRIGGER_START_INDEX, 0, 0), 0)
            # Setup Trigger stop index to 0
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SG_TRIGGER_STOP_INDEX, 0, 0), 1)

            # Expect Positive Level on Trigger input 0 - start of Trigger pulse, high level of TTL signal _|¯¯
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_SOURCE_CONDITION, 0, 0),
                               ps.TriggerCondition.POSITIVE_LEVEL)
            # Expect Negative Level on Trigger input 0 - end of Trigger pulse, low level of TTL signal ¯¯|_
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_SOURCE_CONDITION, 1, 0),
                               ps.TriggerCondition.NEGATIVE_LEVEL)
            # Set Trigger logic for both trigger inputs to OR
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_LOGIC_OPERATION, 0, 0),
                               ps.TriggerLogicOperation.OR)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_LOGIC_OPERATION, 1, 0),
                               ps.TriggerLogicOperation.OR)
            # Set Trigger OR and AND mask for both trigger inputs
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_AND_MASK, 0, 0), 1)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_OR_MASK, 0, 0), 1)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_AND_MASK, 1, 0), 2)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_OR_MASK, 1, 0), 2)

            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SG_TRIGGER_AUTO_RESET_MODE, 0, 0),
                               si.ENABLED)
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SG_TRIGGER_POST_FRAME_COUNT, 0, 0), 0)

            # Set streaming mode to Trigger
            si.SetProperty_i32(self.handle, si.EPK(si.Property.STREAMING_MODE, 0, 0), si.StreamingMode.TRIGGERED)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method({"clear_values": "Clear previous streamed values"})
    def wait_for_trigger(self, clear_values: bool = True) -> None:
        """
        Wait for HW trigger.
        clear_values should be set to True, when we are collecting sequence of triggers.
        Set it to False, when it's nth trigger within a sequence.
        :param clear_values: Clear Stream readout attribute
        :return: None
        """
        try:
            self.logger.info("Waiting for trigger")
            clear_values = bool(int(clear_values))
            for channel, source in self.stream_sources:
                if clear_values:
                    self.notify((channel, source, ATTR_STREAM_READOUT), None)
                self.stream_buffers[(channel, source)] = []

            si.SetProperty_i32(self.handle, si.EPK(si.Property.STREAMING_ACTIVE, 0, 0), si.ENABLED)
            self.set_status(STATUS_BUSY)
            Thread(target=self.stream_data_thread).start()
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method()
    def stop_streaming(self) -> None:
        """
        Stop streaming, either HW or SW triggered.
        :return: None
        """
        try:
            self.logger.info("Stopping streaming")
            si.SetProperty_i32(self.handle, si.EPK(si.Property.STREAMING_ACTIVE, 3, 0), si.DISABLED)
            self.set_status(STATUS_IDLE)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method()
    def reset_streaming(self):
        """
        Reset all streaming attributes.
        :return:
        """
        self.logger.info("Reset all streaming attributes")
        try:
            si.ResetStreamingConfiguration(self.handle)
            self.set_value([ATTR_STREAMING, ATTR_FRAMERATE], 1)
            try:
                self.set_value([ATTR_STREAMING, ATTR_FILTER_RATE], 1)
            except:
                pass
            self.set_value([ATTR_STREAMING, ATTR_BUFFERS_COUNT], 2)
            self.set_value([ATTR_STREAMING, ATTR_FRAME_AGGREGATION], 1)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method()
    def save_adjustment_results(self):
        self.logger.info(f"Saving adjustment results")
        try:
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.AF_ADJUSTMENT_RESULT_SAVE, 1, 0), 1)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    @expose_method()
    def load_adjustment_results(self):
        self.logger.info(f"Loading adjustment results")
        try:
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.AF_ADJUSTMENT_RESULT_LOAD, 1, 0), 1)
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_full_access(self, value):
        self.logger.info(f"Set full access to {value}")
        try:
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SYS_FULL_ACCESS_CONNECTION, 0, 0), int(value))
        except si.Error as e:
            raise DeviceError(self._get_exception(e))

    def set_pilot_laser(self, value):
        self.logger.info(f"Set pilot laser to {value}")
        try:
            si.SetProperty_i32(self.handle, si.EPK(ps.Property.SYS_PILOT_LASER_ACTIVE, 0, 0), int(value))
        except si.Error as e:
            self.get_pilot_laser()
            raise DeviceError(self._get_exception(e))

    @staticmethod
    def _get_exception(exception):
        """
        Helper method to obtain Exception info from PicoScale.
        :param exception: thrown exception
        :type exception: Exception
        :return: Exception message
        :rtype: str
        """
        err_name = "(0x{:04X})".format(exception.code)
        if exception.code in set(err.value for err in si.ErrorCode):
            err_name = si.ErrorCode(exception.code).name + " " + err_name
        elif exception.code in set(err.value for err in ps.ErrorCode):
            err_name = ps.ErrorCode(exception.code).name + " " + err_name
        return f"SI {exception.func} error: {err_name}."

    def process_buffer(self, buffer) -> None:
        """
        Processes the contents of a buffer received from the API.
        This function would typically e.g. store the data to disc.
        In this example the data is stored in a buffer for later
        processing.
        """
        for source_index in range(buffer.info.numberOfSources):
            values = si.CopyBuffer(self.handle, buffer.info.bufferId, source_index)
            channel, source = self.stream_sources[source_index]
            self.stream_buffers[(channel, source)] += values
        si.ReleaseBuffer(self.handle, buffer.info.bufferId)
        return values == []

    def stream_data_thread(self) -> None:
        """
        Continuously acquire buffers and write those into main big buffer.
        After stream is finished set stream_buffer for channel.
        To explain how this work.
        We are streaming at desired frequency, can be many kHz.
        We wait for event to read buffer and write it into main buffer.
        Important to notice it timeout variable.
        At the beginning we wait for TIMEOUT_INFINITE value, since there is timeout between start and actual start of
        stream.
        Normally, after stream stopped or was aborted, we would stop reading.
        But since we are dealing with buffers, we must be sure, that all data was read.
        For that reason we decrease timeout to 100 ms and wait until timeout.
        Which then is our signal, that we read all the data and can stop the Thread.
        :return:
        """
        with self.readout_lock:
            self.logger.info("Starting streaming and buffer acquisition")
            timeout = si.TIMEOUT_INFINITE
            # This flag signals, that last buffer was empty and we read all buffers
            last_buffer_empty = False
            stream_stopped = False
            while True:
                try:
                    # Wait for new event from PicoScale
                    ev = si.WaitForEvent(self.handle, timeout)
                    if ev.type == si.EventType.STREAMBUFFER_READY:
                        # get buffer data
                        buffer = si.AcquireBuffer(self.handle, ev.bufferId)
                        last_buffer_empty = self.process_buffer(buffer)
                        if last_buffer_empty and stream_stopped:
                            break
                    elif ev.type in (si.EventType.STREAM_STOPPED, si.EventType.STREAM_ABORTED):
                        # Stream was stopped or aborted
                        stream_stopped = True
                        if last_buffer_empty:
                            # Stop reading, only when last buffer was empty
                            break
                        # Reduce readout timeout
                        timeout = 100
                        continue
                    else:
                        print(f"Received unexpected event type: {ev.type} (parameter: {ev.devEventParameter})")
                        break
                except si.bindings.Error:
                    # Here readout was stopped due to readout timeout
                    break

            # Flush main stream buffers
            for channel, source in self.stream_sources:
                # print(len(self.stream_buffers[(channel, source)]))
                self.notify((channel, source, ATTR_STREAM_READOUT), self.stream_buffers[(channel, source)])

    def close(self):
        """
        To properly close PicoScale, we need to close handle.
        Otherwise we cannot connect to Device again.
        """
        try:
            if self.handle is not None:
                si.Close(self.handle)
        except Exception as e:
            pass
        finally:
            self.handle = None
        Device.close(self)

    def get_full_access_state(self) -> None:
        """
        Obtain full access.
        this is important for setting triggering attributes.
        :return:
        """
        full_access = si.GetProperty_i32(self.handle, si.EPK(ps.Property.SYS_FULL_ACCESS_CONNECTION, 0, 0))
        self.set_raw_value(ATTR_FULL_ACCESS, full_access)

    def get_pilot_laser(self) -> None:
        """
        Get status of pilot laser, which is used for alignment.
        :return:
        """
        pilot_laser = si.GetProperty_i32(self.handle, si.EPK(ps.Property.SYS_PILOT_LASER_ACTIVE, 0, 0))
        self.set_raw_value(ATTR_PILOT_LASER, pilot_laser)

    def get_system_stable(self) -> None:
        """
        Get system stability flag.
        If system is not stable, we can't start streaming.
        :return:
        """
        is_stable = si.GetProperty_i32(self.handle, si.EPK(ps.Property.SYS_IS_STABLE, 0, 0))
        self.set_value(ATTR_IS_STABLE, is_stable)

    def get_trigger_state(self) -> None:
        """
        Get state of streaming.
        :return:
        """
        trigger_state = si.GetProperty_i32(self.handle, si.EPK(ps.Property.TRIGGER_STATE, 0, 0))

    def poller_interval(self) -> None:
        """
        At each poller tick, we execute all polled functions.
        :return:
        """
        for fun in self.polled_functions:
            fun()

    def start_polling(self):
        self.polling = True
        self.polled_functions = [self.get_system_stable, self.get_full_access_state]
        self.polling_timer = PreciseCallbackTimer(300, self.poller_interval)
        self.polling_timer.start()

    def stop_polling(self):
        self.polling = False
        self.polling_timer.stop()


class DevicePicoscaleChannel(DeviceChannel):
    polling = False
    polling_timer = None

    def __init__(self, device, channel, device_id=None, config=None):
        self.polled_functions = []
        self.source_groups = {}
        self.use_data_sources = config.get("use_data_sources", [])
        DeviceChannel.__init__(self, device, channel, device_id, config)

    def _init_attributes(self):
        DeviceChannel._init_attributes(self)
        self.create_attribute(ATTR_SOURCES_COUNT, readonly=True)
        self.create_attribute(ATTR_HEAD_TYPE, default_type=self.device.available_head_types.keys(),
                              set_function=self.set_head_type)
        self.create_attribute(ATTR_BEAM_INTERRUPT_TOLERANCE, unit="%", min_value=0, max_value=100,
                              set_function=self.set_beam_interrupt_tolerance, default_type=np.uint8)
        self.create_attribute(ATTR_CHANNEL_ENABLED, readonly=False, set_function=self.set_channel_enabled,
                              default_type=bool)

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        number_of_data_sources = si.GetProperty_i32(self.device.handle,
                                                    si.EPK(si.Property.NUMBER_OF_DATA_SOURCES, self.channel, 0))
        self.set_value(ATTR_SOURCES_COUNT, number_of_data_sources)
        head_type = si.GetProperty_i32(self.device.handle,
                                       si.EPK(ps.Property.CH_HEAD_TYPE, self.channel, 0))
        head_type_index, head_type_category = head_type.to_bytes(2, "little")
        head_type = self.device.available_head_types.inv[(head_type_category, head_type_index)]
        self.set_raw_value(ATTR_HEAD_TYPE, head_type)
        beam_interrupt_tolerance = si.GetProperty_i32(self.device.handle,
                                                      si.EPK(ps.Property.CH_BEAM_INTERRUPT_TOLERANCE, self.channel, 0))
        self.set_raw_value(ATTR_BEAM_INTERRUPT_TOLERANCE, beam_interrupt_tolerance)
        ch_enabled = bool(si.GetProperty_i32(self.device.handle, si.EPK(ps.Property.CH_ENABLED, self.channel, 0)))
        self.set_raw_value(ATTR_CHANNEL_ENABLED, ch_enabled)
        self._configure_sources()
        _finish_configuration()

    def set_channel_enabled(self, value):
        try:
            si.SetProperty_i32(self.device.handle, si.EPK(ps.Property.CH_ENABLED, self.channel, 0),
                               int(value))
        except si.Error as e:
            raise DeviceError(self.device._get_exception(e))

    def set_beam_interrupt_tolerance(self, value):
        try:
            si.SetProperty_i32(self.device.handle, si.EPK(ps.Property.CH_BEAM_INTERRUPT_TOLERANCE, self.channel, 0),
                               int(value))
        except si.Error as e:
            raise DeviceError(self.device._get_exception(e))

    def set_head_type(self, value):
        head_type_category, head_type_index = self.device.available_head_types[value]
        lb = head_type_index.to_bytes(1, "little")
        hb = head_type_category.to_bytes(1, "little")
        header_type = int.from_bytes(lb + hb, "little")
        try:
            si.SetProperty_i32(self.device.handle, si.EPK(ps.Property.CH_HEAD_TYPE, self.channel, 0), header_type)
        except si.Error as e:
            raise DeviceError(self.device._get_exception(e))

    def _configure_sources(self):
        """
        Every channel has multiple Sources.
        We first obtain number of Source available on Channel.
        Then get through all of them and filter ones, that are in global variable use_data_source_types.
        :return:
        """
        self.source_groups.clear()
        prefix = {-12: u"p", -9: u"n", -6: u"u", -3: u"m"}

        for source_index in range(self.get_value(ATTR_SOURCES_COUNT)):
            ds_type = si.GetProperty_i32(self.device.handle,
                                         si.EPK(si.Property.DATA_SOURCE_TYPE, self.channel, source_index))
            if ds_type in use_data_source_types or source_index in self.use_data_sources:
                data_type = si.GetProperty_i32(self.device.handle,
                                               si.EPK(si.Property.DATA_TYPE, self.channel, source_index))
                base_unit = si.GetProperty_i32(self.device.handle,
                                               si.EPK(si.Property.BASE_UNIT, self.channel, source_index))
                base_resolution = si.GetProperty_i32(self.device.handle,
                                                     si.EPK(si.Property.BASE_RESOLUTION, self.channel,
                                                            source_index))
                streamable = si.GetProperty_i32(self.device.handle,
                                                si.EPK(si.Property.IS_STREAMABLE, self.channel, source_index))
                group_name = f"{use_data_source_types.get(ds_type, f'Source_{source_index}')}"
                self.source_groups[source_index] = group_name
                si_unit = si_units.get(base_unit, None)
                base_unit = None
                if si_unit is not None:
                    base_unit = prefix.get(base_resolution, "") + si_unit
                self.create_attribute(ATTR_VALUE_READOUT, default_type=np.float64,
                                      unit=base_unit, group=group_name, readonly=True, decimals=10)
                self.create_attribute(ATTR_ID, group=group_name, default_value=source_index, readonly=True,
                                      default_type=np.uint16)
                if bool(streamable):
                    streaming_enabled = si.GetProperty_i32(self.device.handle,
                                                           si.EPK(si.Property.STREAMING_ENABLED, self.channel,
                                                                  source_index))
                    self.create_attribute(ATTR_ALLOW_STREAM, default_value=streaming_enabled, group=group_name,
                                          default_type=bool,
                                          set_function=lambda v, source_index=source_index: self.set_allow_streaming(
                                              source_index, v))
                else:
                    self.create_attribute(ATTR_ALLOW_STREAM, default_value=False, group=group_name,
                                          default_type=bool)
                if bool(streamable):
                    self.create_attribute(ATTR_STREAM_READOUT, default_value=[], default_type=TYPE_ARRAY,
                                          group=group_name, readonly=True, streaming_enabled=False,
                                          unit=base_unit, display=False)

    def subject_update(self, key, value, subject):
        DeviceChannel.subject_update(self, key, value, subject)
        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            _, source, attribute = key
            group = self.source_groups[source]
            if attribute == ATTR_STREAM_READOUT:
                if value is not None:
                    self.get_value([group, attribute]).append(value)
                else:
                    self.set_value([group, attribute], [])
            else:
                self.set_value([group, attribute], value)

    def get_readout(self, source_group, attribute):
        if self.device.handle is None:
            return

        source_id = self.get_value([source_group, ATTR_ID])
        readout = si.GetValue_f64(self.device.handle, self.channel, source_id)
        self.set_value([source_group, attribute], readout)
        return readout

    def set_allow_streaming(self, source_index, value):
        si.SetProperty_i32(self.device.handle, si.EPK(si.Property.STREAMING_ENABLED, self.channel, source_index), value)
        max_fa = si.GetProperty_i32(self.device.handle, si.EPK(si.Property.MAX_FRAME_AGGREGATION, 0, 0))
        max_fr = si.GetProperty_i32(self.device.handle, si.EPK(si.Property.MAX_FRAME_RATE, 0, 0))
        self.device.set_attribute([GROUP_STREAMING, ATTR_FRAME_AGGREGATION, MAX], max_fa)
        self.device.set_attribute([GROUP_STREAMING, ATTR_FRAMERATE, MAX], max_fr)
        if value:
            self.device.stream_sources.append((self.channel, source_index))
        else:
            self.device.stream_sources.remove((self.channel, source_index))

    def poller_interval(self):
        for fun in self.polled_functions:
            try:
                fun()
            except si.bindings.Error:
                pass

    def start_polling(self):
        self.polling = True
        self.polled_functions = []
        for group in self.source_groups.values():
            self.polled_functions.append(
                lambda group=group: self.get_readout(group, ATTR_VALUE_READOUT)
            )
        self.polling_timer = PreciseCallbackTimer(300, self.poller_interval)
        self.polling_timer.start()

    def stop_polling(self):
        self.polling = False
        self.polling_timer.stop()

    @expose_method({"position": "New position"})
    def set_position(self, position):
        self.logger.info(f"Reset position to {position}")
        position = int(position)
        si.SetProperty_i64(self.device.handle, si.EPK(ps.Property.CH_POSITION, self.channel, 0), position)



from PyQt5.QtCore import QSize
from PyQt5.QtGui import QIcon, QFont
from PyQt5.QtWidgets import QMainWindow, QPushButton

import kamzik3
from kamzik3.gui.deviceDebugWidget import DeviceDebugWidget
from kamzik3.gui.templates.sessionWindowTemplate import Ui_MainWindow
from kamzik3.snippets.snippetsWidgets import show_centralized_top
from kamzik3.snippets.snippetsYaml import YamlSerializable


class SessionWindow(Ui_MainWindow, QMainWindow, YamlSerializable):

    def __init__(self, config=None):
        super(SessionWindow, self).__init__()
        self.session = kamzik3.session
        self.config = config
        self.tools = []
        self.setupUi(self)

    def setupUi(self, MainWindow):
        super().setupUi(MainWindow)

        index = 0
        columns = 2 if self.config is None else self.config.get("columns", 2)
        if self.config is not None:
            for index, (title, tool, icon) in enumerate(self.config.get("tools", [])):
                column = index % columns
                row = int(index / columns)
                self.add_tool(title, tool, row, column, icon)

        index += 1
        column = index % columns
        row = int(index / columns)
        self.add_tool("Session", DeviceDebugWidget(self.session.device_id), row, column,
                      self.config.get("icon", ":/icons/icons/kamzik.png"))
        self.setWindowIcon(QIcon(self.config.get("icon", ":/icons/icons/kamzik.png")))
        self.setWindowTitle(self.session.device_id)

    def show_tool(self, flag, tool):
        show_centralized_top(tool)

    def add_tool(self, title, tool, row, column, icon=None):
        self.tools.append(tool)
        tool_button = QPushButton(" {}".format(title))
        self.layout_tool_buttons.addWidget(tool_button, row, column)
        if icon is not None:
            tool_icon = QIcon(icon)
            tool_button.setIcon(tool_icon)
            tool.setWindowIcon(tool_icon)
        tool.setWindowTitle(title)
        tool_button.setIconSize(QSize(25, 25))
        tool_button.setStyleSheet("QPushButton {padding:5px}")
        tool_button.setMinimumWidth(150)
        button_font = QFont("Lucida Sans", 8)
        button_font.setBold(True)
        tool_button.setFont(button_font)
        tool_button.clicked.connect(lambda flag, tool=tool: self.show_tool(flag, tool))

    def closeEvent(self, *args, **kwargs):
        for tool in self.tools:
            tool.hide()
            tool.close()

        self.session = None
        self.tools = []
        QMainWindow.closeEvent(self, *args, **kwargs)



import time

from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceScanner(Device):
    last_scan_count = None
    class_filter = None
    attribute_filter = None
    method_filter = None

    def __init__(self, device_id=None, config=None):
        Device.__init__(self, device_id, config)
        self.connect()

    def handle_configuration(self):
        start_at = time.time()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    @expose_method()
    def get_scanner_attributes(self):
        raise NotImplementedError

    @expose_method()
    def get_scanner_macro(self, scanner_input, scanner_attributes, parent_macro=None):
        raise NotImplementedError



import os
import sys

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QMessageBox, QInputDialog, QLineEdit, QDesktopWidget, QAbstractSpinBox, QSpinBox, \
    QDoubleSpinBox, QApplication, QPlainTextEdit


def clear_layout(self, layout):
    """
    Clear given QLayout of all widgets.
    Perform close method on every widget.
    :param self: QWidget
    :param layout: QLayout
    :return: None
    """
    if layout is not None:
        while layout.count():
            item = layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.close()
                widget.deleteLater()
            else:
                clear_layout(self, item.layout())


def show_error_message(message, parent=None):
    """
    Show error message to user.
    :param message: string
    :param parent: QWidget
    :return: None
    """
    error_message = QMessageBox(parent)
    error_message.setIcon(QMessageBox.Critical)
    error_message.setText(u"{}".format(message))
    error_message.open()


def show_info_message(message, parent=None):
    """
    Show error message to user.
    :param message: string
    :param parent: QWidget
    :return: None
    """
    error_message = QMessageBox(parent)
    error_message.setIcon(QMessageBox.Information)
    error_message.setText(u"{}".format(message))
    error_message.open()


def show_question_dialog(question, title="", parent=None, cancel=False):
    """
    Show dialog with two options. Yes or No.
    Return True or False.
    :param question:
    :param title:
    :param parent:
    :return: bool
    """
    if cancel:
        buttons = QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel
    else:
        buttons = QMessageBox.Yes | QMessageBox.No
    answer = QMessageBox.question(parent, title, question, buttons=buttons, defaultButton=QMessageBox.No)
    if answer in (QMessageBox.Yes, QMessageBox.Cancel):
        return answer
    else:
        return False


def show_prompt_dialog(prompt, title="", parent=None):
    """
    Show prompt dialog to the user.
    Return user input when Enter is pressed.
    Return False if user cancelled prompt dialog.
    :param prompt: string
    :param title: string
    :param parent: QWidget
    :return: string
    """
    user_comment, answer = QInputDialog.getText(parent, title, prompt, QLineEdit.Normal, "")

    if answer:
        return user_comment
    else:
        return False


def show_password_dialog(prompt, title="Type password", default_text="", parent=None):
    """
    Ask user for password input.
    :param prompt: str
    :param title: str
    :param default_text: str
    :param parent: QWidget
    :return: str
    """
    answer, ok = QInputDialog.getText(parent, title, prompt, QLineEdit.Password, default_text)
    if ok:
        return answer


def show_centralized_top(widget):
    """
    Show widget in center of current screen.
    Put widget on top of others.
    Make it active even if minimized.
    :param widget: QWidget
    :return: None
    """
    fg = widget.frameGeometry()
    cp = QDesktopWidget().availableGeometry().center()
    fg.moveCenter(cp)
    widget.move(fg.topLeft())
    widget.setWindowState(Qt.WindowActive)
    widget.show()
    widget.setFocus()
    widget.activateWindow()
    widget.raise_()


def init_qt_app(enable_hd_scaling=False):
    """
    Initialize QApplication with or without hd scaling factor
    :param enable_hd_scaling: bool
    :return: QApplication
    """
    if enable_hd_scaling:
        os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
    else:
        os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "0"
        os.environ["QT_SCALE_FACTOR"] = "1"
        os.environ["QT_SCREEN_SCALE_FACTOR"] = "1"

    app = QApplication(sys.argv)
    if enable_hd_scaling:
        app.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    return app


class SpinBoxFeatures(QAbstractSpinBox):
    returnPressed = pyqtSignal("PyQt_PyObject")

    def __init__(self, reset_on_blur=False, parent=None):
        self.reset_on_blur = reset_on_blur
        QAbstractSpinBox.__init__(self, parent)

    def update_value(self):
        return None

    def keyPressEvent(self, event):
        """
        Set value only when user hit Enter or Return
        :param event:
        :return:
        """
        if event.key() in (Qt.Key_Enter, Qt.Key_Return):
            self.returnPressed.emit(self.value())
            self.lineEdit().deselect()
        elif event.key() == Qt.Key_Comma:
            return
        return QAbstractSpinBox.keyPressEvent(self, event)

    def wheelEvent(self, QWheelEvent):
        """
        Prevent value change by scroll mouse wheel
        :param QWheelEvent:
        :return:
        """
        return None

    def focusOutEvent(self, QFocusEvent):
        """
        Set last known value when widget is out of focus
        :param QFocusEvent:
        :return:
        """
        if self.reset_on_blur:
            self.update_value()
            self.lineEdit().deselect()
        return QAbstractSpinBox.focusOutEvent(self, QFocusEvent)


class CustomSpinBox(QSpinBox, SpinBoxFeatures):
    returnPressed = pyqtSignal("PyQt_PyObject")

    def update_value(self):
        return None


class CustomDoubleSpinBox(QDoubleSpinBox, SpinBoxFeatures):
    returnPressed = pyqtSignal("PyQt_PyObject")

    def update_value(self):
        return None


class CustomLineEdit(QLineEdit):

    def __init__(self, reset_on_blur=False, parent=None):
        self.reset_on_blur = reset_on_blur
        QLineEdit.__init__(self, parent)

    def update_value(self):
        return None

    def focusOutEvent(self, QFocusEvent):
        """
        Set last known value when widget is out of focus
        :param QFocusEvent:
        :return:
        """
        if self.reset_on_blur:
            self.update_value()
            # self.lineEdit().deselect()
        return QLineEdit.focusOutEvent(self, QFocusEvent)


class CustomTextEdit(QPlainTextEdit):

    def __init__(self, reset_on_blur=False, parent=None):
        self.reset_on_blur = reset_on_blur
        QPlainTextEdit.__init__(self, parent)

    def update_value(self):
        return None

    def focusOutEvent(self, QFocusEvent):
        """
        Set last known value when widget is out of focus
        :param QFocusEvent:
        :return:
        """
        if self.reset_on_blur:
            self.update_value()
        return QPlainTextEdit.focusOutEvent(self, QFocusEvent)



from PyQt5.QtCore import pyqtSignal, pyqtSlot, Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QPushButton, QLabel, QSizePolicy, QWidget, QHBoxLayout, QLineEdit, QSpacerItem, QGridLayout, \
    QFrame
from kamzik3.devices.deviceClient import DeviceClient

from kamzik3 import DeviceClientError, CommandFormatException, DeviceError, units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.gui.attributeDeviceDisplayWidget import AttributeDeviceDisplayWidget
from kamzik3.gui.attributeDisplayWidget import AttributeDisplayWidget
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.templates.deviceDebugTemplate import Ui_Form
from kamzik3.snippets.snippetsWidgets import clear_layout, show_error_message


class DeviceDebugWidget(Ui_Form, DeviceWidget):
    sigNewCommand = pyqtSignal("QString")
    command_terminator = None

    def __init__(self, device=None, config=None, parent=None):
        self.groups = []
        DeviceWidget.__init__(self, device, config=config, parent=parent)
        if not isinstance(self.device, DeviceClient):
            self.button_reconnect.hide()

    def setupUi(self, form):
        super(DeviceDebugWidget, self).setupUi(form)
        self.button_connect.hide()
        self.button_enable.hide()
        self.label_device_id.setText(self.device_id)
        self.button_enable.setEnabled(False)
        self.button_disable.setEnabled(False)
        self.button_terminal.setEnabled(False)
        self.page_3.setEnabled(False)
        self.setWindowTitle(u"{} debug".format(self.device_id))
        self.setWindowIcon(QIcon(":/icons/icons/settings.png"))

    def init_signals(self):
        super().init_signals()
        self.sigNewCommand.connect(self.output_commands.setPlainText)

    def keyPressEvent(self, key_event):
        if key_event.key() == Qt.Key_Escape and hasattr(self.device, "stop"):
            self.device.stop()

        DeviceWidget.keyPressEvent(self, key_event)

    @pyqtSlot()
    def slot_handle_configuration(self):
        DeviceWidget.slot_handle_configuration(self)
        if len(self.groups) == 0:
            self.generate_attribute_layout()
            self.generate_methods_layout()

        self.device.attach_attribute_callback(ATTR_ENABLED, self.on_enabled_changed)

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        DeviceWidget.slot_set_status(self, value)
        if value in READY_DEVICE_STATUSES:
            self.button_disconnect.show()
            self.button_disconnect.setEnabled(True)
            self.button_connect.hide()

            self.button_enable.setEnabled(True)
            self.button_disable.setEnabled(True)
            self.button_terminal.setEnabled(True)
            self.page.setEnabled(True)
            self.page_2.setEnabled(True)
            self.page_3.setEnabled(True)
            self.page_4.setEnabled(True)
            self.slot_set_enabled(self.device.get_value(ATTR_ENABLED))
            self.button_reconnect.hide()
        else:
            self.button_enable.setEnabled(False)
            self.button_disable.setEnabled(False)
            self.button_terminal.setEnabled(False)
            self.page.setEnabled(False)
            self.page_2.setEnabled(False)
            self.page_3.setEnabled(False)
            self.page_4.setEnabled(False)
            self.button_disconnect.hide()
            self.button_connect.show()
            if isinstance(self.device, DeviceClient):
                self.button_reconnect.show()
        if value == STATUS_DISCONNECTED:
            self.button_connect.setEnabled(True)

        self.label_device_id.setStyleSheet("QLabel {{background:{}}}".format(self.get_status_color(value)))

    @pyqtSlot("bool")
    def slot_set_enabled(self, flag):
        self.button_enable.setVisible(not flag)
        self.button_disable.setVisible(flag)
        self.page.setEnabled(flag)
        self.page_2.setEnabled(flag)
        self.page_3.setEnabled(flag)
        self.page_4.setEnabled(flag)

    @pyqtSlot(int)
    def slot_attribute_group_changed(self, page):
        self.tab_attribute_groups.currentWidget().setFocus()

    def show_attributes_widget(self):
        self.logger.debug(u"Showing page with device attributes")
        self.widgets_stack.setCurrentIndex(0)

    def show_configuration_widget(self):
        self.logger.debug(u"Showing page with device configuration")
        self.widgets_stack.setCurrentIndex(1)

    def show_terminal_widget(self):
        self.logger.debug(u"Showing page with device terminal")
        self.widgets_stack.setCurrentIndex(2)
        self.input_command.setFocus()

    def show_methods_widget(self):
        self.logger.debug(u"Showing page with methods")
        self.widgets_stack.setCurrentIndex(3)

    def send_command(self):
        self.input_command.selectAll()
        command = self.input_command.text().strip()
        if self.command_terminator:
            command += self.command_terminator
        self.logger.debug(u"Sending command: {0!r}".format(command))
        try:
            self.device.command(command, callback=self.receive_command)
        except (DeviceClientError, CommandFormatException, DeviceError) as e:
            self.logger.exception(u"Sending command error")
            self.sigNewCommand.emit(str(e))

    def receive_command(self, *args):
        if isinstance(args[-1], list):
            self.sigNewCommand.emit("".join(args[-1]))
        else:
            self.sigNewCommand.emit(args[-1])

    def disconnect_device(self):
        self.logger.debug(u"Disconnecting from device")
        self.device.disconnect()
        self.button_disconnect.setDisabled(True)

    def connect_device(self):
        self.logger.debug(u"Connecting to device")
        self.device.reconnect()
        self.button_connect.setDisabled(True)

    def enable_device(self):
        self.logger.debug(u"Enabling device")
        self.device.set_attribute((ATTR_ENABLED, VALUE), 1)

    def disable_device(self):
        self.logger.debug(u"Disabling device")
        self.device.set_attribute((ATTR_ENABLED, VALUE), 0)

    def generate_methods_layout(self):
        device_methods = self.device.exposed_methods
        layout = self.widget_grid_for_methods.layout()

        def exec_method(method_name, argument_inputs=None):
            kwargs = {}
            for argument_input in argument_inputs:
                if isinstance(argument_input, AttributeDisplayWidget):
                    value = argument_input.get_attribute_value()
                    if isinstance(value, units.Quantity):
                        kwargs[argument_input.name] = "{:~}".format(value)
                    else:
                        kwargs[argument_input.name] = value
                else:
                    kwargs[argument_input.name] = argument_input.text()
            try:
                output = getattr(self.device, method_name)(**kwargs)
            except DeviceError as e:
                show_error_message(e, parent=self)
                output = e
            self.output_methods.setPlainText("{} output:\n'{}'".format(method_name, output))

        if device_methods:
            for row, (method_name, method_attributes) in enumerate(device_methods):

                method_label = QLabel(method_name + " (")
                method_label.setAlignment(Qt.AlignCenter | Qt.AlignRight)
                method_label.setStyleSheet("QLabel {font-weight:bold}")
                layout.addWidget(method_label, row, 0)
                method_execute_button = QPushButton("< Call >")
                method_execute_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
                argument_inputs = []
                if method_attributes:
                    inputs_widget = QWidget()
                    inputs_layout = QHBoxLayout()
                    inputs_widget.setLayout(inputs_layout)
                    for attribute_title, attribute_type in method_attributes.items():
                        device_attribute = self.device.get_attribute(attribute_type)
                        if device_attribute is None:
                            inputs_layout.addWidget(QLabel("{}".format(attribute_title)))
                            input_value = QLineEdit()
                            input_value.setPlaceholderText(attribute_type)
                            input_value.name = attribute_title
                            inputs_layout.addWidget(input_value)
                            argument_inputs.append(input_value)
                        else:
                            attribute_widget = AttributeDisplayWidget(attribute_title, device_attribute.attribute_copy())
                            attribute_widget.name = attribute_title
                            inputs_layout.addWidget(attribute_widget.label_widget)
                            inputs_layout.addWidget(attribute_widget.input_widget)
                            inputs_layout.addWidget(attribute_widget.unit_widget)
                            argument_inputs.append(attribute_widget)
                    layout.addWidget(inputs_widget, row, 1)
                else:
                    label = QLabel("None")
                    label.setAlignment(Qt.AlignCenter | Qt.AlignHCenter)
                    label.setStyleSheet("QLabel {font-weight:bold}")
                    layout.addWidget(label, row, 1)

                method_execute_button.released.connect(
                        lambda method_name=method_name, argument_inputs=argument_inputs: exec_method(method_name,
                                                                                                 argument_inputs))
                method_label = QLabel(")")
                method_label.setStyleSheet("QLabel {font-weight:bold}")
                layout.addWidget(method_label, row, 2)
                layout.addWidget(method_execute_button, row, 3)

            spacer = QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)
            layout.addItem(spacer, row + 1, 2)
        else:
            label = QLabel("No methods available")
            label.setAlignment(Qt.AlignCenter | Qt.AlignHCenter)
            label.setStyleSheet("QLabel {font-size:15px; font-weight:bold}")
            layout.addWidget(label, 0, 0)

    def generate_attribute_layout(self):
        group = self._generate_attribute_group(u"General")
        self._generate_attribute_groups(self.device.attributes, group, attribute_path=[])

    def _reset_attribute_layout(self):
        while self.groups:
            self.tab_attribute_groups.removeTab(0)
            clear_layout(self.groups[0], self.groups[0].layout())
            self.groups[0].deleteLater()
            del self.groups[0]
        self.groups = []

    def _generate_attribute_groups(self, attributes, group, attribute_path):
        column, row = 0, 0
        columns = 4 if self.config is None else self.config.get("columns", 4)
        is_picture_in = False
        for key, value in attributes.items():
            if Attribute.is_attribute(value):
                if value.get(DISPLAY) == False:
                    continue
                if column % columns == 0:
                    row += 1
                    column = 0
                attribute_widget = self._generate_attribute_element(attribute_path + [key])
                attribute_widget.setParent(group)
                group.layout().addWidget(attribute_widget, row, column)
                column += 1
                if value.get(DISPLAY) == IMAGE:
                    is_picture_in = True
            else:
                new_group = self._generate_attribute_group(key)
                self._generate_attribute_groups(value, new_group, attribute_path + [key])
        if not is_picture_in:
            spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)
            group.layout().addItem(spacer, row + 1, 0)

        row_count = group.layout().rowCount()
        column_count = group.layout().columnCount()
        for col in range(column_count - 1):
            line = QFrame()
            line.setFrameShape(QFrame.VLine)
            line.setFrameShadow(QFrame.Plain)
            line.setStyleSheet("QFrame {color: #f0f0f0; padding-left: 5px; padding-right: 5px}")
            group.layout().addWidget(line, 1, col + 1, row_count, 1)

    def _generate_attribute_group(self, group_title):
        group_widget = QWidget()
        layout = QGridLayout(group_widget)
        layout.setParent(group_widget)
        layout.setContentsMargins(9, 9, 9, 9)
        layout.setSpacing(9)
        self.groups.append(group_widget)
        self.tab_attribute_groups.insertTab(len(self.groups) - 1, group_widget, group_title)
        return group_widget

    def _generate_attribute_element(self, attribute_path):
        config = {
            "label_width": 110,
        }
        attribute_widget = AttributeDeviceDisplayWidget(self.device_id, attribute=attribute_path, config=config)
        return attribute_widget

    def set_command_terminator(self, terminator):
        if terminator == u"None":
            self.command_terminator = None
        elif terminator == u"\\r\\n":
            self.command_terminator = u"\r\n"
        elif terminator == u"\\r":
            self.command_terminator = u"\r"
        elif terminator == u"\\n":
            self.command_terminator = u"\n"
        elif terminator == u"\\t":
            self.command_terminator = u"\t"

    def filter_attributes(self):
        filter_string = self.input_attribute_filter.text()
        attributes_layout = self.tab_attribute_groups.currentWidget().layout()
        for widget_index in range(attributes_layout.count()):
            widget = attributes_layout.itemAt(widget_index).widget()
            if isinstance(widget, AttributeDeviceDisplayWidget):
                if widget.title.find(filter_string) != -1:
                    widget.show()
                else:
                    widget.hide()

    def close(self):
        self.device.detach_attribute_callback(ATTR_ENABLED, self.on_enabled_changed)
        self.tab_attribute_groups.blockSignals(True)
        self._reset_attribute_layout()
        return DeviceWidget.close(self)

    def reconnect_device(self):
        try:
            self.device.server_reconnect()
        except Exception as e:
            show_error_message(e, parent=self)


import argparse
import os
import signal

import oyaml as yaml

import kamzik3
from kamzik3.snippets.snippetsWidgets import init_qt_app

if __name__ == "__main__":
    # Parse arguments from commandline
    parser = argparse.ArgumentParser()
    parser.add_argument("--conf", help="Path to configuration file in yaml format",
                        default="./conf_dummy_server.yml")
    parser.add_argument("--chdir", help="Path to active directory",
                        default="./")
    args = parser.parse_args()

    # Set active directory
    os.chdir(args.chdir)
    # Create PyQT5 APPLICATION
    app = init_qt_app(enable_hd_scaling=False)

    # Load configuration file
    with open(args.conf, "r") as configFile:
        config = yaml.load(configFile, Loader=yaml.Loader)
    # Start control loop for devices specified in configuration file
    kamzik3.session.set_config(args.conf)
    kamzik3.session.start_control_loops()

    signal.signal(signal.SIGINT, signal.SIG_DFL)
    if "session_window" in config:
        config["session_window"].show()

    # Start PyQT5 app loop
    app.exec_()

    # Close session
    kamzik3.session.stop()



import tempfile
from os import path

import pytest

from kamzik3.devices.deviceSession import DeviceSession


@pytest.fixture(scope="session", autouse=True)
def deviceSession():
    with tempfile.TemporaryDirectory() as root_log:
        session_config = {'attributes': {('Log directory', 'Value'): path.join(root_log, "server_log"),
                                         ('Resource directory', 'Value'): path.join(root_log, "resources"),
                                         ('Allow attribute log', 'Value'): False}}
    session = DeviceSession(device_id="Device_Session", config=session_config)
    yield session
    session.stop()



import json
import logging
import os
import re
import time
from math import ceil
from threading import Lock

import pandas as pd
from PyQt5.QtCore import QAbstractTableModel, Qt, pyqtSlot, QModelIndex, pyqtSignal, QItemSelection
from PyQt5.QtGui import QBrush, QColor, QKeySequence
from PyQt5.QtWidgets import QWidget, QColorDialog, QHeaderView, QShortcut, QButtonGroup, QPushButton, \
    QSpacerItem, QSizePolicy
from natsort import humansorted

import kamzik3
from kamzik3 import DeviceError, SEP, session, units
from kamzik3.constants import *
from kamzik3.devices.deviceSavedAttributes import UPDATE_INSERT_GROUPS, UPDATE_REMOVE_GROUP, UPDATE_INSERT_ROW, \
    UPDATE_VALUE, UPDATE_REMOVE_ROW
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.templates.savedAttributesTemplate import Ui_Form
from kamzik3.snippets.snippetsUnits import device_units
from kamzik3.snippets.snippetsWidgets import show_question_dialog, show_prompt_dialog, show_error_message


class ViewModel(QAbstractTableModel):
    sig_value_update = pyqtSignal(int, int, str)
    colors = {}

    def __init__(self, model_data, from_index=0, to_index=0, parent=None):
        self.model_data = model_data
        self.from_index = from_index
        self.to_index = to_index
        self.previous_table = None
        QAbstractTableModel.__init__(self, parent)

    def columnCount(self, parent=None, *args, **kwargs):
        columns = self.model_data.columns
        return len(columns) - 1

    def rowCount(self, parent=None, *args, **kwargs):
        display = self.to_index - self.from_index
        return display

    def data(self, index, role=None):
        row, column = index.row(), index.column()
        row += self.from_index + 1
        if role == Qt.DisplayRole:
            if column <= 1:
                return "   " + self.model_data.loc[row][column + 1] + "   "
            else:
                return self.model_data.loc[row][column + 1]
        elif role == Qt.BackgroundRole:
            color = self.model_data.loc[row][0]
            if color not in self.colors:
                self.colors[color] = QBrush(QColor(color))
            return self.colors[color]
        elif role == Qt.EditRole:
            current_value = self.model_data.loc[row][column + 1]
            return current_value

    def headerData(self, p_int, Qt_Orientation, role=None):
        if role == Qt.DisplayRole and Qt_Orientation == Qt.Horizontal:
            data = self.model_data.columns[p_int + 1]
            return data
        elif role == Qt.DisplayRole and Qt_Orientation == Qt.Vertical:
            return " " + str(self.from_index + p_int + 1) + " "

    def setData(self, index, value, role=None):
        if role == Qt.EditRole:
            row, column = index.row(), index.column()
            self.sig_value_update.emit(self.from_index + row + 1, column + 1, value)
            return True
        return False

    def flags(self, index):
        if index.column() == 1:
            return Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled

        return Qt.ItemIsSelectable | Qt.ItemIsEnabled


class SavedAttributesWidget(Ui_Form, DeviceWidget):
    default_row_color = "#ffffff"
    file_path = None
    page_size = 100
    sig_refresh_group_tables = pyqtSignal(str)
    sig_refresh_group_table = pyqtSignal(bool, str, object)
    displayed_group_path = None
    column_pattern = re.compile("(.*) \[(.*)\]\[(.*)\]")

    def __init__(self, device=None, config=None, parent=None):
        self.rows_selection_mode = False
        self.button_group = QButtonGroup()
        self.update_lock = Lock()
        self.cache_h5_file = os.path.join(kamzik3.session.get_value(ATTR_CACHE_DIRECTORY), "client_save.h5")
        self.cache_h5_content = None
        self.groups = {}
        self.logger = logging.getLogger("Gui.Device.SavedAttributeWidget")

        if device is None:
            QWidget.__init__(self, parent)
        else:
            DeviceWidget.__init__(self, device, config=config, parent=parent)

        self.manipulation_buttons = [self.button_set_value, self.button_change_color, self.button_remove]
        for button in self.manipulation_buttons:
            button.setDisabled(True)

        self.preset_group_buttons = (self.combo_preset, self.button_remove_preset_group, self.button_add_preset_group)
        self.sig_refresh_group_tables.connect(self.slot_refresh_group_tables)
        self.sig_refresh_group_table.connect(self.slot_refresh_group_table)

        self.shortcut_x = QShortcut(QKeySequence("Alt+X"), self)
        self.shortcut_x.activated.connect(self.set_attribute_value)

        self.shortcut_a = QShortcut(QKeySequence("Alt+A"), self)
        self.shortcut_a.activated.connect(self.add_row)

        self.shortcut_d = QShortcut(QKeySequence("Alt+D"), self)
        self.shortcut_d.activated.connect(self.remove_rows)

        self.shortcut_c = QShortcut(QKeySequence("Alt+C"), self)
        self.shortcut_c.activated.connect(self.select_color)

        self.table_attributes.setSelectionBehavior(self.table_attributes.SelectItems)
        self.table_attributes.setSelectionMode(self.table_attributes.SingleSelection)
        self.table_attributes.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.table_attributes.horizontalHeader().setStretchLastSection(True)
        self.table_attributes.doubleClicked.connect(self.on_double_click)

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        if value in READY_DEVICE_STATUSES and not self.configured:
            self.configured = False
            self.slot_handle_configuration()
            self.device.attach_attribute_callback(ATTR_LAST_UPDATE, self.content_update, key_filter=VALUE)
            with open(self.cache_h5_file, "wb") as h5_fp:
                h5_fp.write(self.device.get_value(ATTR_CONTENT))
            self.device.attach_attribute_callback(ATTR_GROUPS, self.groups_changed, key_filter=VALUE)
            self.configured = True

    @pyqtSlot(object)
    def groups_changed(self, groups):
        if groups is None:
            return

        selected_item = self.combo_preset.currentText()
        self.combo_preset.clear()
        self.groups = {}
        # Add only top groups to selection
        for group in groups:
            group_path = group.split("/")[1:]
            if group_path[0] not in self.groups:
                self.groups[group_path[0]] = [group_path[1]]
            else:
                self.groups[group_path[0]].append(group_path[1])

        # We want alphabetically sorted group names
        sorted_groups = humansorted(self.groups.keys())
        self.combo_preset.addItems(sorted_groups)
        item_index = self.combo_preset.findText(selected_item)

        # Select group which was previously selected
        if item_index != -1:
            self.combo_preset.setCurrentIndex(item_index)
        else:
            self.combo_preset.setCurrentIndex(0)

    def slot_handle_configuration(self):
        pass

    @pyqtSlot()
    def content_update(self, value):
        if value is None or not self.configured:
            return
        with self.update_lock:
            values = value.split(SEP)
            if values[0] == UPDATE_INSERT_GROUPS:
                groups_data = json.loads(values[1])
                for group, data in groups_data.items():
                    df = pd.DataFrame(data, columns=list(data.keys()))
                    df.to_hdf(self.cache_h5_file, key=group, format="table", complevel=1)
            elif values[0] == UPDATE_REMOVE_GROUP:
                with pd.HDFStore(self.cache_h5_file) as hdf:
                    for group_path in self.device.get_value(ATTR_GROUPS):
                        if group_path.startswith(f"/{values[1]}/"):
                            hdf.remove(group_path)
            elif values[0] == UPDATE_INSERT_ROW:
                group = values[1]
                df = pd.read_hdf(self.cache_h5_file, key=group, mode="a")
                data = json.loads(values[2])
                position = int(values[3])
                line = pd.DataFrame(data, index=[position])
                df2 = pd.concat([df.iloc[:position - 1], line, df.iloc[position - 1:]])
                df2.reset_index(drop=True, inplace=True)
                df2.to_hdf(self.cache_h5_file, group, format="fixed")
                subgroup = values[1].split("/")
                if subgroup[1] == self.combo_preset.currentText() and subgroup[
                    2] == self.button_group.checkedButton().text():
                    self.sig_refresh_group_table.emit(True, subgroup[-1], df2)
            elif values[0] == UPDATE_VALUE:
                group = values[1]
                df = pd.read_hdf(self.cache_h5_file, key=group, mode="a")
                df.loc[int(values[2]), df.columns[int(values[3])]] = values[4]
                df.to_hdf(self.cache_h5_file, group, format="fixed")
                subgroup = values[1].split("/")
                if subgroup[1] == self.combo_preset.currentText() and subgroup[
                    2] == self.button_group.checkedButton().text():
                    self.sig_refresh_group_table.emit(True, subgroup[-1], df)
            elif values[0] == UPDATE_REMOVE_ROW:
                group = values[1]
                df = pd.read_hdf(self.cache_h5_file, key=group, mode="a")
                df.drop(labels=int(values[2]), axis=0, inplace=True)
                df.reset_index(drop=True, inplace=True)
                df.to_hdf(self.cache_h5_file, group, format="fixed")
                subgroup = values[1].split("/")
                if subgroup[1] == self.combo_preset.currentText() and subgroup[
                    2] == self.button_group.checkedButton().text():
                    self.sig_refresh_group_table.emit(True, subgroup[-1], df)

    @pyqtSlot(str)
    def slot_refresh_group_table(self, group):
        self.preset_group_changed(self.combo_preset.currentText())

    @pyqtSlot(int, int, str)
    def slot_value_update(self, row, column, value):
        top_group = self.combo_preset.currentText()
        for sub_group in self.groups[top_group]:
            group_path = "/" + "/".join((top_group, sub_group))
            self.device.update_value(group=group_path, row=row, column=column, value=value)

    @pyqtSlot(QItemSelection, QItemSelection)
    def current_selection_changed(self, selected, deselected):
        if self.rows_selection_mode:
            for button in self.manipulation_buttons + [self.button_add]:
                button.setDisabled(True)
        else:
            self.button_add.setDisabled(False)
            for button in self.manipulation_buttons:
                button.setDisabled(True if len(selected) == 0 else False)

            if len(selected) > 0:
                if selected[0].topLeft().column() <= 1:
                    self.button_set_value.setDisabled(True)

    @pyqtSlot(QModelIndex)
    def on_double_click(self, index):
        selected_index = self.table_attributes.selectedIndexes()[0]
        if self.checkbox_allow_double_click.isChecked() and selected_index.column() > 0:
            self.set_attribute_value()

    @pyqtSlot()
    def set_attribute_value(self):
        row, column = self.get_selected_cell()
        if column is None or column <= 1:
            return

        attribute_value = self.table_attributes.model().model_data.iloc[row, column]
        column_header = self.table_attributes.model().model_data.columns[column]
        result = self.column_pattern.search(column_header)
        if result is not None:
            try:
                device_id, attribute_name, unit = result.group(1), result.group(2), result.group(3)
                self.logger.info(f"Set {device_id} {attribute_name} to value {attribute_value} {unit}")
                device = session.get_device(device_id)
                attribute_as_list = eval(f"[{attribute_name}]")
                value_in_device_units = device_units(device, attribute_as_list, units.Quantity(attribute_value, unit))
                device.set_value(attribute_as_list, value_in_device_units.m)
            except (DeviceError, Exception) as e:
                error_message = f"Cannot set {device_id} {attribute_name} to value {attribute_value} {unit}.\r\n{e}"
                show_error_message(error_message, self)

    def get_selected_cell(self):
        selected_index = self.table_attributes.selectedIndexes()
        if not selected_index:
            return None, None
        row = selected_index[0].row()

        if row is not None:
            row_offset = self.table_attributes.model().from_index
            row = row_offset + selected_index[0].row() + 1
            return row, selected_index[0].column() + 1
        else:
            return None, None

    @pyqtSlot()
    def remove_rows(self):
        with self.update_lock:
            row, column = self.get_selected_cell()
            if column is None:
                return

            answer = show_question_dialog(f"Do You really want to remove row {row} ?", "Confirmation", self)
            if answer:
                top_group = self.combo_preset.currentText()
                self.logger.info(f"Removing row {row} from group {top_group}")
                for sub_group in self.groups[top_group]:
                    group_path = "/" + "/".join((top_group, sub_group))
                    self.device.remove_row(group=group_path, row=row)

    @pyqtSlot()
    def add_row(self):
        with self.update_lock:
            user_comment = show_prompt_dialog("Type Your comment:", "Add new row", self)
            if user_comment:
                top_group = self.combo_preset.currentText()
                self.logger.info(f"Prepending new row in group {top_group}")
                for subgroup in self.groups[top_group]:
                    group_path = os.path.join("/", top_group, subgroup)
                    df = pd.read_hdf(self.cache_h5_file, key=group_path)
                    table_row = ["#fff", time.strftime("%d-%m-%Y %H:%M:%S"), user_comment]
                    for column in df.columns[3:]:
                        result = self.column_pattern.search(column)
                        if result is not None:
                            try:
                                device_id, attribute_name, unit = result.group(1), result.group(2), result.group(3)
                                attribute_as_list = eval(f"[{attribute_name}]")
                                device = session.get_device(device_id)
                                current_value = device.get_attribute(attribute_as_list).value().to(unit)
                                table_row.append(current_value.m)
                            except (DeviceError, Exception):
                                table_row.append(None)
                        else:
                            table_row.append(None)
                    table_row = [str(v) for v in table_row]
                    self.device.insert_row(group=group_path, data=dict(zip(df.columns, table_row)), position=2)

    @pyqtSlot()
    def select_color(self):
        with self.update_lock:
            row, column = self.get_selected_cell()
            if row is None:
                return

            color = QColorDialog().getColor()
            if color is not None and color.name() != "#000000":
                self.slot_value_update(row, 0, str(color.name()))

    @pyqtSlot()
    def add_preset_group(self):
        with self.update_lock:
            group_name = show_prompt_dialog("Set name of a new group", "New save group")
            if group_name and group_name != "":
                self.logger.info(f"Adding new group {group_name}")
                groups = {}
                top_group = self.combo_preset.currentText()
                for subgroup in self.groups[top_group]:
                    group_path = os.path.join("/", top_group, subgroup)
                    new_group_path = os.path.join("/", group_name, subgroup)
                    df = pd.read_hdf(self.cache_h5_file, key=group_path)
                    groups[new_group_path] = dict(zip(df.columns, [["None"]] * len(df.columns)))
                self.device.insert_groups(groups_data=groups)

    @pyqtSlot()
    def remove_preset_group(self):
        with self.update_lock:
            current_group = self.combo_preset.currentText()
            answer = show_question_dialog(f"Do You really want to remove group {current_group} ?", "Confirmation", self)
            if answer:
                self.logger.info(f"Removing group {current_group}")
                self.device.remove_group(group=current_group)

    @pyqtSlot("QString")
    def slot_refresh_group_tables(self, top_group):
        """
        Preset group was changed.
        We need to update attribute table.
        @param top_group: name of top group
        @type top_group: string
        """
        if top_group == "":
            return

        self.input_page.blockSignals(True)
        self.input_page.setValue(1)
        self.input_page.blockSignals(False)

        while self.layout_subgroups_select.count() != 0:
            self.layout_subgroups_select.removeItem(self.layout_subgroups_select.itemAt(0))
        self.button_group = QButtonGroup()
        for button_index, subgroup in enumerate(self.groups[top_group]):
            button = QPushButton(subgroup)
            button.setStyleSheet("QPushButton {background-color:silver;} QPushButton:checked {background-color:#f6b442;font-weight:bold}")
            button.setCheckable(True)
            button.setAutoExclusive(True)
            button.setMinimumHeight(30)
            button.toggled.connect(lambda state, subgroup=subgroup: self.slot_refresh_group_table(state, subgroup))
            self.button_group.addButton(button, button_index)
            self.layout_subgroups_select.addWidget(button)
        self.layout_subgroups_select.addSpacerItem(QSpacerItem(0, 0, hPolicy=QSizePolicy.Expanding))
        self.button_group.button(0).setChecked(True)
        self.check_page_borders()

    pyqtSignal(bool, str)
    pyqtSignal(bool, str, object)

    def slot_refresh_group_table(self, state, subgroup, model_data=None):
        if not state:
            return

        row_offset = (self.input_page.value() - 1) * self.page_size
        top_group = self.combo_preset.currentText()
        with self.update_lock:
            if model_data is None:
                model_data = pd.read_hdf(self.cache_h5_file, "/".join([top_group, subgroup]))
            row_count = len(model_data.index) - 1
            to_index = (row_offset + self.page_size) if row_count >= (row_offset + self.page_size) else row_count
            view_model = ViewModel(model_data, from_index=row_offset, to_index=to_index)
            view_model.sig_value_update.connect(self.slot_value_update)
            self.table_attributes.setModel(view_model)
            self.table_attributes.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeToContents)
            self.table_attributes.horizontalHeader().setSectionResizeMode(1, QHeaderView.Interactive)
            self.table_attributes.selectionModel().selectionChanged.connect(self.current_selection_changed)
            self.input_page.blockSignals(True)
            pages_count = ceil(row_count / self.page_size)
            self.input_page.setMaximum(pages_count if pages_count > 0 else 1)
            self.input_page.blockSignals(False)

    @pyqtSlot()
    def slot_jump_to_first_page(self):
        self.input_page.setValue(1)

    @pyqtSlot()
    def slot_jump_to_last_page(self):
        self.input_page.setValue(self.input_page.maximum())

    @pyqtSlot("int")
    def slot_jump_to_page(self, page_index):
        self.slot_refresh_group_table(True, self.button_group.checkedButton().text())
        self.check_page_borders()

    def check_page_borders(self):
        page_index = self.input_page.value()
        at_max = page_index == self.input_page.maximum()
        at_min = page_index == 1
        self.input_page.setSuffix(f" / {self.input_page.maximum()}")
        self.button_next_page.setDisabled(at_max)
        self.button_last_page.setDisabled(at_max)
        self.button_previous_page.setDisabled(at_min)
        self.button_first_page.setDisabled(at_min)



import inspect
from collections import OrderedDict

import oyaml as yaml
from yaml import MappingNode, ScalarNode, CollectionNode, SequenceNode

class ConfigHolder():

    def __init__(self, suffix, **kwargs):
        self.suffix = suffix
        self.conf = kwargs

class YamlDumper(yaml.Dumper):

    def represent_data(self, data):
        """
        Custom Yaml representer.
        Every class inheriting from YamlSerializable is serialized using this representer.
        Use traditional serialization on other classes.
        :param data: mixed
        :return: Yaml Node
        """
        device_py_path = "{}.{}".format(data.__class__.__module__, data.__class__.__name__)
        if isinstance(data, YamlSerializable):
            if self.ignore_aliases(data):
                self.alias_key = None
            else:
                self.alias_key = id(data)
            if self.alias_key is not None:
                if self.alias_key in self.represented_objects:
                    node = self.represented_objects[self.alias_key]
                    # if node is None:
                    #    raise RepresenterError("recursive objects are not allowed: %r" % data)
                    return node
                # self.represented_objects[alias_key] = None
                self.object_keeper.append(data)

            output_node = self.represent_mapping(u"!Device:{}".format(device_py_path), data.yaml_mapping(),
                                                 flow_style=False)
            # self.anchor_node(output_node)
            return output_node
        elif isinstance(data, ConfigHolder):
            if self.ignore_aliases(data):
                self.alias_key = None
            else:
                self.alias_key = id(data)
            if self.alias_key is not None:
                if self.alias_key in self.represented_objects:
                    node = self.represented_objects[self.alias_key]
                    return node
                # self.represented_objects[alias_key] = None
                self.object_keeper.append(data)
            output_node = self.represent_mapping(data.suffix, data.conf, flow_style=False)
            # self.anchor_node(output_node)
            return output_node
        else:
            try:
                return yaml.Dumper.represent_data(self, data)
            except TypeError:
                return yaml.Dumper.represent_none(self, "")

class YamlConfigSerializer(yaml.Loader):
    def pass_device(self, suffix, node):
        arguments = yaml.Loader.construct_mapping(self, node, deep=True)
        mapping_node = MappingNode(u'tag:yaml.org,2002:python/object{}'.format(":kamzik3.snippets.snippetsYaml.ConfigHolder"), [])
        new_device = yaml.Loader.construct_object(self, mapping_node)
        new_device.__init__("!Device"+suffix, **arguments)
        return new_device

    def pass_widget(self, suffix, node):
        arguments = yaml.Loader.construct_mapping(self, node, deep=True)
        mapping_node = MappingNode(u'tag:yaml.org,2002:python/object{}'.format(":kamzik3.snippets.snippetsYaml.ConfigHolder"), [])
        new_device = yaml.Loader.construct_object(self, mapping_node)
        new_device.__init__("!Widget"+suffix, **arguments)
        return new_device

YamlConfigSerializer.add_multi_constructor(u'!Device', YamlConfigSerializer.pass_device)
YamlConfigSerializer.add_multi_constructor(u'!Widget', YamlConfigSerializer.pass_widget)

class YamlSerializable(object):

    def yaml_mapping(self):
        """
        Returns mapping of all arguments for __init__ function.
        If You want to return custom Yaml mapping re-implement this method.
        :return: OrderedDict
        """
        arg_spec = inspect.signature(self.__init__).parameters.keys()
        mapping = OrderedDict()
        for arg in arg_spec:
            try:
                argument_value = self.__getattribute__(arg)
                if argument_value is not None:
                    mapping[arg] = argument_value
            except (KeyError, AttributeError):
                """
                Key does not exists, skip in that case
                If You want to implement missing argument, re-implement yaml_mapping in
                particular device class.
                """
                pass
        return mapping



import time

from kamzik3.devices.general.deviceGrabber import DeviceGrabber

try:
    import cv2
except ModuleNotFoundError:
    raise Exception("Module cv2 (opencv-python) was not found.")
import numpy as np

from kamzik3.constants import *


class DeviceCv2(DeviceGrabber):
    api_presets = {
        "Any": cv2.CAP_ANY,
        "FFmpeg": cv2.CAP_FFMPEG,
        "Gstreamer": cv2.CAP_GSTREAMER,
    }

    def _init_attributes(self):
        DeviceGrabber._init_attributes(self)
        self.create_attribute(ATTR_SOURCE)
        self.create_attribute(ATTR_SOURCE_FRAMERATE, default_value=0, default_type=np.float, unit="FPS", min_value=0,
                              max_value=10000, decimals=2, readonly=True)
        self.create_attribute(ATTR_SOURCE_WIDTH, default_value=0, default_type=np.uint, unit="px", min_value=0,
                              max_value=2 ** 32, readonly=True)
        self.create_attribute(ATTR_SOURCE_HEIGHT, default_value=0, default_type=np.uint, unit="px", min_value=0,
                              max_value=2 ** 32, readonly=True)
        self.create_attribute(ATTR_AUTO_SET_FRAMERATE, default_value=True, default_type=bool,
                              set_function=self.set_auto_frameset)
        self.create_attribute(ATTR_CAPTURE_API, default_value="Any", default_type=self.api_presets.keys())

    def _acquire_loop(self):
        self.stopped.clear()
        capture_api = self.api_presets.get(self.get_value(ATTR_CAPTURE_API), cv2.CAP_ANY)
        try:
            cap = cv2.VideoCapture(int(self.get_value(ATTR_SOURCE)), capture_api)
        except ValueError:
            cap = cv2.VideoCapture(self.get_value(ATTR_SOURCE), capture_api)

        if not cap.isOpened():
            cap.release()
            return
        self.set_value(ATTR_SOURCE_FRAMERATE, cap.get(cv2.CAP_PROP_FPS))
        self.set_value(ATTR_SOURCE_WIDTH, cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.set_value(ATTR_SOURCE_HEIGHT, cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        if self.get_value(ATTR_AUTO_SET_FRAMERATE):
            self.set_value(ATTR_FRAMERATE, self.get_value(ATTR_SOURCE_FRAMERATE))
        self.frame_delay = 1e9 / self.get_value(ATTR_FRAMERATE)
        self.set_status(STATUS_BUSY)
        fps_counter_reference = time_reference = time.perf_counter_ns()
        grabbed_frames_count = 0
        fps_counter = 0
        while not self.stopped.wait((self.frame_delay - (time.perf_counter_ns() - time_reference)) * 1e-9):
            time_reference = time.perf_counter_ns()
            ret, cv_img = cap.read()
            grabbed_frames_count += 1
            fps_counter += 1
            if ret:
                self.set_value(["Grabber", "Frame"], cv2.cvtColor(cv_img, cv2.COLOR_BGR2RGB))

            if (time.perf_counter_ns() - fps_counter_reference) >= 1e9:
                self.set_raw_value(ATTR_FRAMERATE, fps_counter)
                self.set_raw_value(ATTR_FRAMES_GRABBED, grabbed_frames_count)
                fps_counter = 0
                fps_counter_reference = time.perf_counter_ns()

    def set_auto_frameset(self, value):
        if value:
            self.set_value(ATTR_FRAMERATE, self.get_value(ATTR_SOURCE_FRAMERATE))



import copy

import numpy as np
import pytest
from numpy.testing import assert_allclose

from kamzik3 import units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute

@pytest.fixture(
    params=[
        {"attribute_id": ATTR_POSITION, "default_value": 8.25, "min_value": -10, "max_value": 100, "unit": "nm",
         "default_type": np.float, "decimals": 2, "readonly": False, "description": "Motor position",
         "tolerance": (3, 5), "offset": 0, "factor": 1},
    ])
def attribute_params(request):
    return request.param


@pytest.fixture(params=[pytest.lazy_fixture("attribute_params")])
def attribute(request):
    yield Attribute(**request.param)


def test_attribute_creation(attribute, attribute_params):
    assert isinstance(attribute, Attribute)
    assert attribute.attribute_id == ATTR_POSITION
    assert attribute[ATTR_DESCRIPTION] == attribute_params["description"]
    unit = attribute_params["unit"]
    assert attribute.value() == units.Quantity(f"{attribute_params['default_value']} {unit}")
    assert attribute.within_limits(attribute.value())
    assert attribute.unit() == unit
    assert attribute.negative_tolerance() == units.Quantity(f"{attribute_params['tolerance'][0]} {unit}")
    assert attribute.positive_tolerance() == units.Quantity(f"{attribute_params['tolerance'][1]} {unit}")
    assert attribute.minimum() == units.Quantity(f"{attribute_params['min_value']} {unit}")
    assert attribute.maximum() == units.Quantity(f"{attribute_params['max_value']} {unit}")


@pytest.mark.parametrize("factor", [-1, 1])
@pytest.mark.parametrize("offset", [0, 10, -15])
@pytest.mark.parametrize("value", [-10, 80, 0])
def test_offset_factor_value(attribute, attribute_params, factor, offset, value):
    _attribute = copy.deepcopy(attribute)
    _attribute[OFFSET] = offset
    _attribute[FACTOR] = factor
    unit = attribute_params["unit"]

    if _attribute.numerical:
        assert _attribute.apply_offset_factor(value) == (value + offset) * factor
        assert _attribute.apply_offset_factor(-value) == (-value + offset) * factor
        assert _attribute.remove_offset_factor(value) == (value / factor) - offset
        assert _attribute.remove_offset_factor(-value) == (-value / factor) - offset

        _attribute[VALUE] = units.Quantity(f"{value} {unit}")
        assert _attribute.remove_offset_factor() == (_attribute.value().m / factor) - offset
        assert _attribute.within_limits(_attribute.value())
        assert _attribute.setpoint() == units.Quantity(f"{value} {unit}")
        assert _attribute.within_tolerance(_attribute.value())
